import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import minimize

from .PointHomogeneous import PointHomogeneous


class MiniBall:
    def __init__(self, points: list, metric: str = "euclidean"):
        """
        Initialize the MiniBall class
        :param points: np.ndarray - array of points in the space
        """
        self.points = points
        self.metric = metric

        self.number_of_points = len(self.points)
        self.dimension = self.points[0].coordinates.size
        self.center = np.zeros(self.dimension)
        self.radius = 10.0

        self.optimization_results = self.get_ball()
        self.center = PointHomogeneous(self.optimization_results.x[: self.dimension])
        self.radius = self.optimization_results.x[self.dimension]

    def get_ball(self):
        """
        Find the smallest ball containing all given points in Euclidean metric
        """

        def objective_function(x):
            """
            Objective function to minimize the squared radius r^2 of the ball
            """
            return x[-1] ** 2

        # Prepare constraint equations based on the metric
        if self.metric == "hofer":

            def constraint_equations(x):
                """
                For Hofer metric, constraint equations must satisfy the ball by:
                r - radius of the sphere, x - one of given points,
                c - center of the sphere
                """
                return NotImplemented

        else:

            def constraint_equations(x):
                """
                For Euclidean metric, constraint equations must satisfy the ball by:
                r^2 - (x - c)^2 >= 0
                r - radius of the sphere, x - one of given points,
                c - center of the sphere
                """
                constraints = np.zeros(self.number_of_points)
                for i in range(self.number_of_points):
                    # in case of Euclidean metric, the normalized point has to be taken
                    # in account
                    squared_distance = sum(
                        (self.points[i].normalize()[j] - x[j]) ** 2
                        for j in range(self.dimension)
                    )
                    constraints[i] = x[-1] ** 2 - squared_distance
                return constraints

        # Prepare inequality constraint dictionary
        ineq_con = {"type": "ineq", "fun": constraint_equations}

        # Initialize optimization variables
        initial_guess = np.zeros(self.dimension + 1)
        initial_guess[0] = 1.0
        initial_guess[-1] = self.radius

        # Perform optimization
        result = minimize(objective_function, initial_guess, constraints=ineq_con)
        return result

    def plot_ball(self, ax=None, color="red", alpha=0.15):
        """
        Plot the ball in 3D space
        :param ax: matplotlib.axes.Axes object
        :param color: color of the ball
        :param alpha: transparency of the ball
        """
        # check dimension
        if not self.dimension == 4:
            raise ValueError(
                "Cannot plot ball with dimension other than 3 (4 in homogenoeus coordinates)"
            )
        else:
            if ax is None:
                ax = plt.figure().add_subplot(projection="3d")
            else:
                ax = ax

            # Create the 3D sphere representing the circle
            u = np.linspace(0, 2 * np.pi, 100)
            v = np.linspace(0, np.pi, 100)
            x = (
                self.radius * np.outer(np.cos(u), np.sin(v))
                + self.center.normalized_in_3d()[0]
            )
            y = (
                self.radius * np.outer(np.sin(u), np.sin(v))
                + self.center.normalized_in_3d()[1]
            )
            z = (
                self.radius * np.outer(np.ones(np.size(u)), np.cos(v))
                + self.center.normalized_in_3d()[2]
            )
            ax.plot_surface(x, y, z, color=color, alpha=alpha)

        return ax



import numpy as np

from .DualQuaternion import DualQuaternion
from .PointHomogeneous import PointHomogeneous
from .RationalCurve import RationalCurve


class AffineMetric:
    """
    Class of affine metric in R12

    For more information, see M. Hofer's dissertation thesis titled "Variational Motion
    Design in the Presence of Obstacles", section 2.2 on page 6.
    """
    def __init__(self, motion_curve: RationalCurve, points: list[PointHomogeneous]):
        """
        Construct the affine metric of a motion from the given points in the 3D space

        :param motion_curve: RationalCurve - rational curve representing the motion
        :param points: list[PointHomogeneous] - points in the 3D space
        """
        self.motion_curve = motion_curve
        self.points = points
        self.number_of_points = len(points)

        self.matrix = self.create_affine_metric()

    def __repr__(self):
        return f"{self.matrix}"

    def create_affine_metric(self) -> np.ndarray:
        """
        Create the affine metric of the motion

        This function computes the metric matrix for a homogeneous 3D point based on
        the formulation from M. Hofer's dissertation thesis titled "Variational Motion
        Design in the Presence of Obstacles", specifically on page 7, equation 2.4.

        :return: np.ndarray - affine metric matrix in R12x12

        :references:
            - M. Hofer, "Variational Motion Design in the Presence of Obstacles",
            dissertation thesis (2004), Page 7, Equation 2.4
        """
        metric_matrix = np.zeros((12, 12))
        for i in range(self.number_of_points):
            metric_matrix += self.get_point_metric_matrix(self.points[i])
        return metric_matrix

    @staticmethod
    def get_point_metric_matrix(point: PointHomogeneous) -> np.ndarray:
        """
        Get the metric matrix of the given point

        :param point: PointHomogeneous - point in the 3D space

        :return: np.ndarray - metric matrix of a single point in R12x12

        :references:
            - M. Hofer, "Variational Motion Design in the Presence of Obstacles",
            dissertation thesis (2004), Page 7, Equation 2.4
        """
        p = point.normalized_in_3d()
        i = np.eye(3)

        m00 = i
        m01 = p[0] * i
        m02 = p[1] * i
        m03 = p[2] * i

        m11 = p[0] ** 2 * i
        m12 = p[0] * p[1] * i
        m13 = p[0] * p[2] * i

        m22 = p[1] ** 2 * i
        m23 = p[1] * p[2] * i

        m33 = p[2] ** 2 * i

        metric_matrix = np.block([[m00, m01, m02, m03],
                                  [m01, m11, m12, m13],
                                  [m02, m12, m22, m23],
                                  [m03, m13, m23, m33]])

        return metric_matrix

    def get_curve_transformations(self) -> list[DualQuaternion]:
        """
        Get the transformations of the curve

        :return: list[DualQuaternion] - transformations of the curve
        """

        # tranformation at -1
        dq_1 = self.motion_curve.evaluate(-1)
        # transformation at 1
        dq_2 = self.motion_curve.evaluate(1)
        # transformation at infinity
        dq_inf = self.motion_curve.evaluate(0, inverted_part=True)

        return [DualQuaternion(dq_1), DualQuaternion(dq_2), DualQuaternion(dq_inf)]

    def distance(self, a: np.ndarray, b: np.ndarray) -> float:
        """
        Distance between two affine displacements

        :param a:
        :param b:
        :return: float - distance between a and b
        """
        return np.sqrt(self.inner_product(a - b, a - b))

    def inner_product(self, dq_a: DualQuaternion, dq_b: DualQuaternion):
        """
        Inner product of two DualQuaternions

        :param dq_a: DualQuaternion
        :param dq_b: DualQuaternion

        :return: float - inner product of dq_a and dq_b
        """
        inner_product = np.zeros(4)
        for i in range(self.number_of_points):
            a_point = dq_a.act(self.points[i])
            b_point = dq_b.act(self.points[i])

            point = np.dot(a_point.array(), b_point.array())
            inner_product += point

        return inner_product



from typing import Union

import sympy as sp

from .DualQuaternion import DualQuaternion
from .RationalCurve import RationalCurve
from .RationalDualQuaternion import RationalDualQuaternion
from .TransfMatrix import TransfMatrix


class MotionInterpolation:
    """
    Method for interpolation of poses by rational motion curve in SE(3).

    There are two methods for interpolation of poses by rational motion curve, please
    see the following examples for more details.

    :see also: :ref:`interpolation_background`, :ref:`interpolation_examples`

    :examples:

    .. testcode::

        # 4-pose interpolation

        from rational_linkages import (DualQuaternion, Plotter, FactorizationProvider,
                                       MotionInterpolation, RationalMechanism)


        if __name__ == "__main__":
            # 4 poses
            p0 = DualQuaternion()  # identity
            p1 = DualQuaternion.as_rational([0, 0, 0, 1, 1, 0, 1, 0])
            p2 = DualQuaternion.as_rational([1, 2, 0, 0, -2, 1, 0, 0])
            p3 = DualQuaternion.as_rational([3, 0, 1, 0, 1, 0, -3, 0])

            # obtain the interpolated motion curve
            c = MotionInterpolation.interpolate([p0, p1, p2, p3])

            # factorize the motion curve
            f = FactorizationProvider().factorize_motion_curve(c)

            # create a mechanism from the factorization
            m = RationalMechanism(f)

            # create an interactive plotter object, with 1000 descrete steps
            # for the input rational curves, and arrows scaled to 0.5 length
            myplt = Plotter(interactive=True, steps=1000, arrows_length=0.5)
            myplt.plot(m, show_tool=True)

            # plot the poses
            for pose in [p0, p1, p2, p3]:
                myplt.plot(pose)

            # show the plot
            myplt.show()

    .. testcode::

        # 3-pose interpolation

        from rational_linkages import DualQuaternion, Plotter, MotionInterpolation


        if __name__ == "__main__":
            p0 = DualQuaternion([0, 17, -33, -89, 0, -6, 5, -3])
            p1 = DualQuaternion([0, 84, -21, -287, 0, -30, 3, -9])
            p2 = DualQuaternion([0, 10, 37, -84, 0, -3, -6, -3])

            c = MotionInterpolation.interpolate([p0, p1, p2])

            plt = Plotter(interactive=False, steps=500, arrows_length=0.05)
            plt.plot(c, interval='closed')

            for i, pose in enumerate([p0, p1, p2]):
                plt.plot(pose, label='p{}'.format(i+1))
    """
    def __init__(self):
        """
        Creates a new instance of the rational motion interpolation class.
        """
        pass

    @staticmethod
    def interpolate(poses: list[Union[DualQuaternion, TransfMatrix]]) -> RationalCurve:
        """
        Interpolates the given 3 poses by a rational motion curve in SE(3).

        :param list[Union[DualQuaternion, TransfMatrix]] poses: The poses to
            interpolate.

        :return: The rational motion curve.
        :rtype: RationalCurve
        """
        # check number of poses
        if not (len(poses) == 3 or len(poses) == 4):
            raise ValueError('The number of poses must be 3 or 4.')

        rational_poses = []

        # convert poses to rational dual quaternions
        for pose in poses:
            if isinstance(pose, TransfMatrix):
                rational_poses.append(DualQuaternion.as_rational(pose.matrix2dq()))
            elif isinstance(pose, DualQuaternion) and not pose.is_rational:
                rational_poses.append(DualQuaternion.as_rational(pose.array()))
            elif isinstance(pose, DualQuaternion) and pose.is_rational:
                rational_poses.append(pose)
            else:
                raise TypeError('The given poses must be either TransfMatrix '
                                 'or DualQuaternion.')

        # interpolate the rational poses
        if len(rational_poses) == 4:
            curve_eqs = MotionInterpolation.interpolate_cubic(rational_poses)
            return RationalCurve(curve_eqs)
        else:
            curve_eqs = MotionInterpolation.interpolate_quadratic(rational_poses)
            return RationalCurve(curve_eqs)

    @staticmethod
    def interpolate_quadratic(poses: list[DualQuaternion]) -> list[sp.Poly]:
        """
        Interpolates the given 3 rational poses by a quadratic curve in SE(3).

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: The rational motion curve.
        :rtype: list[sp.Poly]
        """
        alpha = sp.Symbol('alpha')
        omega = sp.Symbol('omega')
        t = sp.Symbol('t')

        p0 = poses[0].array()
        p1 = poses[1].array()
        p2 = poses[2].array()

        c = alpha * p0 + (p1 - alpha * p0 - omega * p2) * t + omega * p2 * t**2

        symbolic_curve = RationalDualQuaternion(c)

        # apply Stydy condition, i.e. obtain epsilon norm of the curve
        study_norm = symbolic_curve.norm()

        # simplify the norm
        study_norm = sp.simplify(study_norm[4] / (t * (t - 1)))

        # obtain the equations for alpha and omega
        eq0 = study_norm.subs(t, 0)
        eq1 = study_norm.subs(t, 1)

        # solve the equations
        sols = sp.solve([eq0, eq1], [alpha, omega])

        # get non zero solution
        nonzero_sol = None
        for sol in sols:
            if (not (sol[0] == 0 and sol[1] == 0)
                    and sol[0].is_number
                    and sol[1].is_number):
                nonzero_sol = sol

        if nonzero_sol is None:
            raise ValueError('Interpolation Failed for the given poses.' 
                             'Tip: if you used *nice* numbers to create poses, try to '
                             'slightly alter them. For example, instead of offset 0.0 '
                             'put 0.0001.')

        al = nonzero_sol[0]
        om = nonzero_sol[1]
        # obtain the interpolated curve
        c_interp = al * p0 + (p1 - al * p0 - om * p2) * t + om * p2 * t**2

        # list of polynomials
        poly = [sp.Poly(el, t) for el in c_interp]

        return poly

    @staticmethod
    def interpolate_cubic(poses: list[DualQuaternion]) -> list[sp.Poly]:
        """
        Interpolates the given 4 rational poses by a cubic curve in SE(3).

        The 4 poses span a projective 3-space, which is intersected with Study quadric.
        This intersection gives another quadric containing all 4 poses, and it also
        contains cubic curves if it contains lines. The algorithm later searches
        for one of the cubic curves that interpolates the 4 poses.

        :see also: :ref:`interpolation_background`

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: The rational motion curve.
        :rtype: list[sp.Poly]

        :raises ValueError: If the interpolation has no solution, 'k' does not exist.
        """
        # obtain additional dual quaternions k1, k2
        try:
            k = MotionInterpolation._obtain_k_dq(poses)
        except Exception:
            raise ValueError('Interpolation has no solution.')

        # solve for t[i] - the parameter of the rational motion curve for i-th pose
        t_sols = MotionInterpolation._solve_for_t(poses, k)

        # Lagrange's interpolation part
        # lambdas for interpolation - scalar multiples of the poses
        lams = sp.symbols("lams1:5")

        parametric_points = [sp.Matrix(poses[0].array()),
                             sp.Matrix(lams[0] * poses[1].array()),
                             sp.Matrix(lams[1] * poses[2].array()),
                             sp.Matrix(lams[2] * poses[3].array())]

        # obtain the Lagrange interpolation for poses p0, p1, p2, p3
        interp = MotionInterpolation._lagrange_poly_interpolation(parametric_points)

        t = sp.symbols("t:4")
        x = sp.symbols("x")

        # to avoid calculation with infinity, substitute t[i] with 1/t[i]
        temp = [element.subs(t[0], 0) for element in interp]
        temp2 = [element.subs(x, 1 / x) for element in temp]
        temp3 = [sp.together(element * x ** 3) for element in temp2]
        temp4 = [sp.together(element.subs({t[1]: 1 / t_sols[0],
                                           t[2]: 1 / t_sols[1],
                                           t[3]: 1 / t_sols[2]}))
                 for element in temp3]

        # obtain additional parametric pose p4
        lam = sp.symbols("lam")
        poses.append(DualQuaternion([lam, 0, 0, 0, 0, 0, 0, 0]) - k[0])

        eqs_lambda = [element.subs(x, lam) - lams[-1] * poses[-1].array()[i]
                      for i, element in enumerate(temp4)]

        sols_lambda = sp.solve(eqs_lambda, lams, domain='RR')

        # obtain the family of solutions
        poly = [element.subs(sols_lambda) for element in temp4]

        # choose one solution by setting lambda, in this case lambda = 0
        poly = [element.subs(lam, 0) for element in poly]

        t = sp.Symbol("t")
        poly = [element.subs(x, t) for element in poly]

        return [sp.Poly(element, t) for element in poly]

    @staticmethod
    def _obtain_k_dq(poses: list[DualQuaternion]) -> list[DualQuaternion]:
        """
        Obtain additional dual quaternions k1, k2 for interpolation of 4 poses.

        :param list[DualQuaternion] poses: The rational poses to interpolate.

        :return: Two additional dual quaternions for interpolation.
        :rtype: list[DualQuaternion]
        """
        x = sp.symbols("x:3")

        k = DualQuaternion(poses[0].array() + x[0] * poses[1].array()
                           + x[1] * poses[2].array() + x[2] * poses[3].array())

        eqs = [k[0], k[4], k.norm().array()[4]]

        sol = sp.solve(eqs, x, domain=sp.S.Reals)

        k_as_expr = [sp.Expr(el) for el in k]

        k1 = [el.subs({x[0]: sol[0][0], x[1]: sol[0][1], x[2]: sol[0][2]})
              for el in k_as_expr]
        k2 = [el.subs({x[0]: sol[1][0], x[1]: sol[1][1], x[2]: sol[1][2]})
              for el in k_as_expr]

        k1_dq = DualQuaternion([el.args[0] for el in k1])
        k2_dq = DualQuaternion([el.args[0] for el in k2])

        return [k1_dq, k2_dq]

    @staticmethod
    def _solve_for_t(poses: list[DualQuaternion], k: list[DualQuaternion]):
        """
        Solve for t[i] - the parameter of the rational motion curve for i-th pose.

        :param list[DualQuaternion] poses: The rational poses to interpolate.
        :param list[DualQuaternion] k: The additional dual quaternions for interpolation.

        :return: The solutions for t[i].
        :rtype: list
        """
        t = sp.symbols("t:3")

        study_cond_mat = sp.Matrix([[0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                                    [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, 1],
                                    [1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                                    [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0]])

        t_dq = [DualQuaternion([t[i], 0, 0, 0, 0, 0, 0, 0]) for i in range(3)]

        eqs = [sp.Matrix((t_dq[0] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[1].array()),
               sp.Matrix((t_dq[1] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[2].array()),
               sp.Matrix((t_dq[2] - k[0]).array()).transpose() @ study_cond_mat
               @ sp.Matrix(poses[3].array())]

        sols_t = sp.solve(eqs, t)

        # covert to list and retrun
        return [val for i, (key, val) in enumerate(sols_t.items())]

    @staticmethod
    def _lagrange_polynomial(degree, index, x, t):
        """
        Calculate the Lagrange polynomial for interpolation.

        :param int degree: The degree of the Lagrange polynomial.
        :param int index: The index of the Lagrange polynomial.
        :param symbol x: The interpolation point (indeterminate).
        :param list[symbol] t: The interpolation nodes.

        :return: The Lagrange polynomial.
        :rtype: sp.Expr
        """
        lagrange_poly = 1
        for i in range(degree + 1):
            if i != index:
                lagrange_poly *= (x - t[i]) / (t[index] - t[i])
        return lagrange_poly

    @staticmethod
    def _lagrange_poly_interpolation(poses: list[sp.Matrix]):
        """
        Calculate the interpolation polynomial using Lagrange interpolation.

        :param list[sp.Matrix] poses: The poses to interpolate.

        :return: The interpolation polynomial.
        :rtype: sp.Matrix
        """
        # indeterminate x
        x = sp.symbols('x')

        # interpolation nodes
        t = sp.symbols("t:4")

        degree = len(poses) - 1
        result = sp.Matrix([0, 0, 0, 0, 0, 0, 0, 0])

        for i in range(degree + 1):
            result += poses[i] * MotionInterpolation._lagrange_polynomial(degree,
                                                                          i, x, t)
        return result



from typing import Union

import numpy as np
import sympy as sp

from .DualQuaternion import DualQuaternion
from .Linkage import Linkage
from .NormalizedLine import NormalizedLine
from .PointHomogeneous import PointHomogeneous
from .RationalCurve import RationalCurve


class MotionFactorization(RationalCurve):
    """
    Class representing Motion Factorization sequence

    Inherits from :class:`rational_linkages.RationalCurve` class. Given as set of polynomials in
    dual quaternion space. You can find more information in the paper by
    :footcite:t:`Frischauf2023`.

    :param list[DualQuaternion] sequence_of_factored_dqs: list of DualQuaternions
        representing the revolute axes of the rational motion factorization

    :ivar list[DualQuaternion] dq_axes: list of DualQuaternions representing the
        revolute axes of the rational motion factorization
    :ivar list[DualQuaternion] factors_with_parameter: parameterized factors of the
        curve
    :ivar int number_of_factors: number of factors of the curve
    :ivar list[Linkage] linkage: list of link-joint connecting points

    :example:

    .. testcode::

        # Motion factorization of a 2R mechanism

        from rational_linkages import DualQuaternion
        from rational_linkages import MotionFactorization


        f1 = MotionFactorization(
            [DualQuaternion([0, 0, 0, 1, 0, 0, 0, 0]),
             DualQuaternion([0, 0, 0, 2, 0, 0, -1, 0])])

    .. footbibliography::

    """

    def __init__(self, sequence_of_factored_dqs: list[DualQuaternion]):
        """
        Initialize a MotionFactorization object

        :param list[DualQuaternion] sequence_of_factored_dqs: list of DualQuaternions
            representing the revolute axes of the rational motion factorization
        """
        curve_polynomials = self.get_polynomials_from_factorization(
            sequence_of_factored_dqs)
        super().__init__(curve_polynomials)
        self.dq_axes = sequence_of_factored_dqs
        self.factors_with_parameter = self.get_symbolic_factors()
        self.number_of_factors = len(self.dq_axes)

        self.linkage = self.get_joint_connection_points()

    def __repr__(self):
        return f"MotionFactorization({self.factors_with_parameter})"

    @staticmethod
    def get_polynomials_from_factorization(factors: list[DualQuaternion]) -> (
            list)[sp.Poly]:
        """
        Construct rational curve from Dual Quaternions equation factors

        :param list[DualQuaternion] factors: list of sympy polynomials representing
            the curve, 1st row is homogeneous coordinate equation

        :return: motion curve using Sympy polynomials
        :rtype: RationalCurve
        """
        t = sp.Symbol("t")

        polynomial_t = DualQuaternion([t, 0, 0, 0, 0, 0, 0, 0])
        polynomials_dq = DualQuaternion()
        for i in range(len(factors)):
            polynomials_dq = polynomials_dq * (polynomial_t - factors[i])

        return [sp.Poly(polynom, t)
                for i, polynom in enumerate(polynomials_dq.array())]

    def get_symbolic_factors(self) -> list[DualQuaternion]:
        """
        Get symbolic factors of the curve with parameter t, in a form (t - factor)

        :return: list of DualQuaternions representing the curve
        :rtype: list[DualQuaternion]
        """
        t = sp.Symbol("t")
        polynomial_t = DualQuaternion([t, 0, 0, 0, 0, 0, 0, 0])
        return [polynomial_t - self.dq_axes[i] for i in range(len(self.dq_axes))]

    def get_numerical_factors(self, t_numerical: float) -> list[DualQuaternion]:
        """
        Get numerical factors of the curve with parameter t, in a form
        (t - dq_axes)

        :param float t_numerical: parameter of the motion curve

        :return: list of numerical DualQuaternions factors of the curve
        :rtype: list[DualQuaternion]
        """
        dq = DualQuaternion([t_numerical, 0, 0, 0, 0, 0, 0, 0])
        return [dq - self.dq_axes[i] for i in range(len(self.dq_axes))]

    def act(
        self, affected_object, param: float, start_idx: int = None, end_idx: int = None
    ):
        """
        Act on an object with the MotionFactorization sequence of given axes

        If the indexes of the axes are not specified, the action is performed
        using all sequence of MotionFactorization axes

        :param PointHomogeneous, NormalizedLine affected_object: object to act on
        :param float param: parameter of the motion curve
        :param int start_idx: index of the first axis to act with
        :param int end_idx: index of the last axis to act with

        :return: object after the action
        :rtype: PointHomogeneous, NormalizedLine
        """
        from .DualQuaternionAction import DualQuaternionAction

        start_idx = 0 if start_idx is None else start_idx
        end_idx = self.number_of_factors - 1 if end_idx is None else end_idx
        acting_sequence = self.get_numerical_factors(param)[start_idx : end_idx + 1]

        action = DualQuaternionAction()
        return action.act(acting_sequence, affected_object)

    def direct_kinematics(self, t_numerical: float, inverted_part: bool = False
                          ) -> list[np.array]:
        """
        Direct kinematics of the rational mechanism

        :param float t_numerical: parameter of the motion curve
        :param bool inverted_part: if True, return the inverted part of the curve

        :return: list of np.array - points of the curve
        :rtype: list[np.ndarray]
        """
        linkage_points = []
        for i in range(self.number_of_factors):
            linkage_points.append(self.linkage[i].points)

        for i in range(self.number_of_factors - 1):
            if inverted_part:
                pts_acted = [self.act(linkage_points[i + 1][j],
                                      end_idx=i, param=1/t_numerical) for j in range(2)]
            else:
                pts_acted = [self.act(linkage_points[i + 1][j],
                                      end_idx=i, param=t_numerical) for j in range(2)]
            linkage_points[i + 1] = pts_acted

        linkage_points = [linkage_points[i][j] for i in range(len(linkage_points))
                          for j in range(len(linkage_points[i]))]

        linkage_points_3d = [np.array(linkage_points[i].normalized_in_3d())
                             for i in range(len(linkage_points))]
        return linkage_points_3d

    def direct_kinematics_of_tool(self, t_numerical: float, end_effector: np.ndarray,
                                  inverted_part=False) -> np.ndarray:
        """
        Direct kinematics of the end effector position

        :param float t_numerical: parameter of the motion curve
        :param np.ndarray end_effector: homogeneous coordinates of the end effector,
            given as np.array([w, x, y, z])
        :param bool inverted_part: if True, return the inverted part of the curve

        :return: list of np.array - point of the tool position
        :rtype: np.ndarray
        """
        ee_point = PointHomogeneous.from_3d_point(end_effector)

        if inverted_part:
            point_after_action = self.act(
                ee_point, end_idx=self.number_of_factors - 1, param=(1 / t_numerical)
            )
        else:
            point_after_action = self.act(
                ee_point, end_idx=self.number_of_factors - 1, param=t_numerical
            )

        end_effector_point = point_after_action.normalized_in_3d()
        return end_effector_point

    def direct_kinematics_of_tool_with_link(self, t_numerical: float,
                                            end_effector: np.ndarray,
                                            inverted_part=False) -> list:
        """
        Direct kinematics of the end effector position and the last link point

        :param float t_numerical: parameter of the motion curve
        :param bool inverted_part: if True, return the inverted part of the curve

        :return: list of np.array - tool and link points
        :rtype: list[np.ndarray]
        """
        ee_point = self.direct_kinematics_of_tool(t_numerical, end_effector,
                                                  inverted_part=inverted_part)
        link_point = self.direct_kinematics(t_numerical,
                                            inverted_part=inverted_part)[-1]

        return [ee_point, link_point]

    def joint_angle_to_t_param(self, joint_angle: Union[np.ndarray, float] = 0,
                               unit: str = 'rad') -> float:
        """
        Convert joint angle to t parameter of the curve

        This method relates the joint rotation angle to the parameter of the rational
        motion curve, i.e. the parameter variable 't'. It uses the rotational quaternion
        of dual quaternion that represents the rotation axis (joint) and reparameterizes
        it by cotangent function. This provides full cycle motion of the joint axis
        from 0 to 2*pi. More information can be found in documentation in `Joint Angle
        to Curve Parameter`_.

        :param float joint_angle: joint angle in radians
        :param str unit: 'rad' or 'deg'

        :return: t parameter of the curve, bool - if True, the inverted part
        :rtype: float

        :seealso: `Joint Angle to Curve Parameter`_

        .. _Joint Angle to Curve Parameter: background-math/joint-angle-to-t.rst
        """
        if unit == 'deg':
            joint_angle = np.deg2rad(joint_angle)
        elif unit != 'rad':
            raise ValueError("unit must be 'rad' or 'deg'")

        normalized_angle = joint_angle % (2 * np.pi)

        # avoid division by zero
        if normalized_angle == 0.0:
            normalized_angle = np.finfo(float).eps

        sqrt_p_norm = np.sqrt(self.dq_axes[0].p.norm())
        t = sqrt_p_norm / np.tan(normalized_angle/2) + self.dq_axes[0].p[0]

        return t

    def t_param_to_joint_angle(self, t_param: float) -> float:
        """
        Convert t parameter of the curve to joint angle

        This is an inverse function of
        :meth:`.MotionFactorization.joint_angle_to_t_param` method. See more
        information in documentation in `Joint Angle to Curve Parameter`_.

        :param float t_param: t parameter of the curve

        :return: joint angle in radians
        :rtype: float
        """
        t_param_joint0 = t_param - self.dq_axes[0].p[0]

        if t_param_joint0 == 0.0:
            t_param_joint0 = np.finfo(float).eps

        angle = 2 * np.arctan(np.sqrt(self.dq_axes[0].p.norm()) / t_param_joint0)

        # normalize angle to [0, 2*pi]
        if angle < 0:
            angle += 2 * np.pi

        return angle

    def factorize(self) -> list['MotionFactorization']:
        """
        Factorize the motion curve into motion factorizations

        :return: list of MotionFactorization objects
        :rtype: list[MotionFactorization]
        """
        from .FactorizationProvider import FactorizationProvider

        factorization_provider = FactorizationProvider()
        return factorization_provider.factorize_for_motion_factorization(self)

    def get_joint_connection_points(self) -> list[Linkage]:
        """
        Get points of the linkage of the MotionFactorization

        :return: list of points of the linkage, the points are the nearest to origin,
            i.e. the foot point of a line (axis)
        :rtype: list[Linkage]
        """
        return [Linkage(axis, [PointHomogeneous.from_3d_point(axis.dq2point_via_line())])
                for axis in self.dq_axes]

    def set_joint_connection_points(self, points: list[PointHomogeneous]):
        """
        Set points of the linkage of the MotionFactorization

        :param list[PointHomogeneous] points: list of points of the linkage
        """
        # pair the input points
        points_pairs = []
        for i in range(len(points) // 2):
            points_pairs.append([points[2 * i], points[2 * i + 1]])

        for i in range(len(points_pairs)):
            self.linkage[i] = Linkage(self.dq_axes[i], points_pairs[i])

    def set_joint_connection_points_by_parameters(self, params: list):
        """
        Set joint connection points based on the given line-parameters.

        :param np.ndarray params: Parameters used to calculate the points
            on the lines.
        """
        for i, linkage in enumerate(self.linkage):
            if len(params[i]) == 1:
                linkage.set_point_by_param(0, params[i][0])
                linkage.set_point_by_param(1, params[i][0])
            elif len(params[i]) == 2:
                linkage.set_point_by_param(0, params[i][0])
                linkage.set_point_by_param(1, params[i][1])
            else:
                raise ValueError("The parameters must be of length 1 or 2.")

    def joint(self, idx: int) -> tuple:
        """
        Returns the joint at the given index.

        :param int idx: The index of the joint

        :return: The joint line and the points of the joint segment
        :rtype: tuple
        """
        point0 = self.linkage[idx].points[0]
        point1 = self.linkage[idx].points[1]

        if np.allclose(point0.normalized_in_3d(), point1.normalized_in_3d()):
            # if the points are the same, add a minimal offset
            min_point = PointHomogeneous(point0.array() + np.array([0, 0, 0, 0.0001]))
            joint = NormalizedLine.from_two_points(point0, min_point)
        else:
            joint = NormalizedLine.from_two_points(point0, point1)

        return joint, point0, point1

    def link(self, idx: int) -> tuple:
        """
        Returns the link at the given index.

        :param int idx: The index of the link

        :return: The link line and the points of the link segment
        :rtype: tuple
        """
        point0 = self.linkage[idx - 1].points[1]
        point1 = self.linkage[idx].points[0]
        link = NormalizedLine.from_two_points(point0, point1)
        return link, point0, point1

    def base_link(self, other_factorization_point: PointHomogeneous) -> tuple:
        """
        Returns the base link.

        :param PointHomogeneous other_factorization_point: The point of the other
            factorization to construct the base link

        :return: The base link line and the points of the base link segment
        :rtype: tuple
        """
        point0 = self.linkage[0].points[0]
        point1 = other_factorization_point

        if np.allclose(point0.normalized_in_3d(), point1.normalized_in_3d()):
            # if the points are the same, add a minimal offset
            point1 = point0 + PointHomogeneous([0, 0, 0, 0.0001])

        link = NormalizedLine.from_two_points(point0, point1)
        return link, point0, point1

    def tool_link(self, other_factorization_point: PointHomogeneous) -> tuple:
        """
        Returns the tool link.

        :param PointHomogeneous other_factorization_point: The point of the other
            factorization to construct the tool link

        :return: The tool link line and the points of the tool link segment
        :rtype: tuple
        """
        point0 = self.linkage[-1].points[1]
        point1 = other_factorization_point

        if np.allclose(point0.normalized_in_3d(), point1.normalized_in_3d()):
            # if the points are the same, add a minimal offset
            point1 = point0 + PointHomogeneous([0, 0, 0, 0.0001])

        link = NormalizedLine.from_two_points(point0, point1)
        return link, point0, point1



from typing import Optional, Sequence, Union
from warnings import warn

import numpy as np
from sympy import Expr, Poly, Symbol, simplify

from .Quaternion import Quaternion

# Forward declarations for class names
NormalizedLine = "NormalizedLine"
PointHomogeneous = "PointHomogeneous"


class DualQuaternion:
    """
    Class representing Dual Quaternions in 3D space.

    Dual Quaternions are used in kinematics and computer graphics for transformations
    and interpolations. They consist of a primal quaternion representing rotation and
    translation and a dual quaternion representing infinitesimal transformations.

    :param list[float] study_parameters: array or list of 8 Study
        parameters. If None, an identity DualQuaternion is constructed.

    :ivar Quaternion p: primal quaternion - the primal part of the Dual Quaternion,
        representing rotation and translation.  See also :class:`~rational_linkages.Quaternion`
    :ivar Quaternion d: dual quaternion - the dual part of the Dual Quaternion,
        representing translation. See also :class:`~rational_linkages.Quaternion`
    :ivar np.ndarray dq: 8-vector of study parameters, representing the Dual Quaternion

    :examples:

    .. testcode::

        # General usage

        from rational_linkages import DualQuaternion
        dq = DualQuaternion([1, 2, 3, 4, 0.1, 0.2, 0.3, 0.4])

    .. testcode::

        # Identity dual quaternion with no rotation, no translation

        from rational_linkages import DualQuaternion
        dq = DualQuaternion()

    .. testcode::

        # Dual quaternion from two quaternions

        from rational_linkages import DualQuaternion
        from rational_linkages import Quaternion
        q1 = Quaternion([0.5, 0.5, 0.5, 0.5])
        q2 = Quaternion([1, 2, 3, 4])
        dq = DualQuaternion.from_two_quaternions(q1, q2)
    """

    def __init__(self, study_parameters: Optional[Sequence[float]] = None):
        """
        Dual Quaternion object, assembled from 8-vector (list or np.array) as DQ,
        or two 4-vectors (np.arrays) as two Quaternions (see @classmethod bellow).
        If no Study's parameters are provided, an identity is constructed.

        :param Optional[Sequence[float]] study_parameters: array or list
            of 8 Study's parameters. If None, an identity DualQuaternion is constructed.
            Defaults to None.
        """
        if study_parameters is not None:
            if len(study_parameters) != 8:
                raise ValueError("DualQuaternion: input has to be 8-vector")
            study_parameters = np.asarray(study_parameters)
            primal = study_parameters[:4]
            dual = study_parameters[4:]
        else:
            primal = np.array([1, 0, 0, 0])
            dual = np.array([0, 0, 0, 0])

        self.p = Quaternion(primal)
        self.d = Quaternion(dual)
        self.dq = self.array()

        # check if all entries of the DQ are rational numbers
        if all(isinstance(x, Expr) for x in self.array()):
            self.is_rational = True
        else:
            self.is_rational = False

    @classmethod
    def from_two_quaternions(
        cls, primal: Quaternion, dual: Quaternion
    ) -> "DualQuaternion":
        """
        Construct DualQuaternion from primal and dual Quaternions.

        :param Quaternion primal: primal part
        :param Quaternion dual: dual part

        :return: DualQuaternion
        :rtype: DualQuaternion
        """
        return cls(np.concatenate((primal.array(), dual.array())))

    @classmethod
    def from_bq_biquaternion(cls, biquaternion):
        """
        Construct DualQuaternion from a biquaternion.

        :param biquaternion_py.biquaternion.BiQuaternion biquaternion: biquaternion

        :return: DualQuaternion
        :rtype: DualQuaternion

        :raises ValueError: if the input is not a
            biquaternion_py.biquaternion.BiQuaternion object

        :examples:

        .. testcode::

            # Construct dual quaternion from a BiQuaternion

            from rational_linkages import DualQuaternion
            from biquaternion_py import II, JJ, KK, EE
            bq = 2*KK + EE * II
            dq = DualQuaternion.from_bq_biquaternion(bq)

        .. testcode::

            # Construct dual quaternion from a BiQuaternion

            from rational_linkages import DualQuaternion
            from biquaternion_py import BiQuaternion
            bq = BiQuaternion(1, 0, 0, 0, 0, 2, 3, 4)
            dq = DualQuaternion.from_bq_biquaternion(bq)
        """
        from biquaternion_py import BiQuaternion

        if not isinstance(biquaternion, BiQuaternion):
            raise ValueError("The input has to be a "
                             "biquaternion_py.biquaternion.BiQuaternion object"
                             "from biquaternion_py package.")

        coeffs = biquaternion.coeffs
        return cls(np.array(coeffs, dtype="float64"))

    @classmethod
    def from_bq_poly(cls, poly, indet):
        """
        Construct DualQuaternion from a biquaternion polynomial.

        The biquaternion polynomial is given in the form (t - h), where t is the
        indeterminant and h is a biquaternion. To obtain the DualQuaternion, the
        coefficients of the polynomial are negated and then assembled into a numpy
        array.

        :param biquaternion_py.polynomials.Poly poly: biquaternion polynomial
        :param sp.Symbol indet: indeterminant of the polynomial

        :return: DualQuaternion
        :rtype: DualQuaternion

        :raises ValueError: if the input is not a biquaternion_py.polynomials.Poly
            object
        :raises ValueError: if the polynomial is not of degree 1

        :examples:

        .. testcode::

            # Construct dual quaternion from a BiQuaternion polynomial

            from rational_linkages import DualQuaternion
            from biquaternion_py import Poly, II, JJ, KK, EE
            from sympy import Symbol

            t = Symbol('t')
            h = 2*KK + EE * II
            poly_bq = Poly(t - h, t)

            # poly_bq must be of form (t - h), i.e. degree 1 otherwise ValueError is raised
            dq = DualQuaternion.from_bq_poly(poly_bq, indet=t)
        """
        from biquaternion_py import Poly

        if not isinstance(poly, Poly):
            raise ValueError("The input has to be a biquaternion_py.polynomials.Poly "
                             "object from biquaternion_py package.")
        elif poly.deg(indet) != 1:
            raise ValueError("The polynomial has to be of degree 1.")

        poly_coeffs = [-x for x in poly.coeff(indet, 0).coeffs]
        return cls(np.array(poly_coeffs, dtype="float64"))

    @classmethod
    def as_rational(cls, study_parameters: Union[list, np.ndarray] = None):
        """
        Assembly of DualQuaternion from Sympy's rational numbers (Rational).

        :param Union[list, np.ndarray] study_parameters: list of 8 numbers

        :return: DualQuaternion with rational elements
        :rtype: DualQuaternion
        """
        from sympy import Rational, nsimplify

        if study_parameters is not None:
            rational_numbers = [nsimplify(x, tolerance=1*(-10)) for x in study_parameters]
        else:
            rational_numbers = [Rational(1), Rational(0), Rational(0), Rational(0),
                                Rational(0), Rational(0), Rational(0), Rational(0)]

        return cls(rational_numbers)

    @classmethod
    def random(cls, interval: float = 1):
        """
        Construct a random DualQuaternion with in the interval (-interval, interval)

        :param float interval: interval of random numbers

        :return: DualQuaternion with random numbers
        :rtype: DualQuaternion
        """
        study_parameters = np.random.uniform(-interval, interval, 8)
        return cls(study_parameters)

    @classmethod
    def random_on_study_quadric(cls, interval: float = 1):
        """
        Construct a random DualQuaternion on the Study quadric

        The interval (-interval, interval) is used to generate random numbers.

        :param float interval: interval of random numbers

        :return: DualQuaternion with random numbers lying on the Study quadric
        :rtype: DualQuaternion
        """
        dq = cls.random(interval)
        return dq.back_projection()

    def __repr__(self):
        """
        Printing method override

        :return: DualQuaterion in readable form
        :rtype: str
        """
        dq = np.array2string(self.array(),
                             precision=10,
                             suppress_small=True,
                             separator=', ',
                             max_line_width=100000)
        return f"{dq}"

    def __getitem__(self, idx) -> np.ndarray:
        """
        Get an element of DualQuaternion

        :param int idx: index of the Quaternion element to call 0..7

        :return: float number of the element
        :rtype: np.ndarray
        """
        element = self.array()
        element = element[idx]  # or, p.dob = p.dob.__getitem__(idx)
        return element

    def __eq__(self, other) -> bool:
        """
        Compare two DualQuaternions if they are equal

        :param DualQuaternion other: DualQuaternion

        :return: True if two DualQuaternions are equal, False otherwise
        :rtype: bool
        """

        return np.array_equal(self.array(), other.array())

    def __add__(self, other) -> "DualQuaternion":
        """
        Addition of two DualQuaternions

        :param DualQuaternion other: other DualQuaternion

        :return: added DualQuaternion
        :rtype: DualQuaternion
        """
        p = self.p + other.p
        d = self.d + other.d
        return DualQuaternion.from_two_quaternions(p, d)

    def __sub__(self, other) -> "DualQuaternion":
        """
        Subtraction of two DualQuaternions

        :param DualQuaternion other: other DualQuaternion

        :return: subtracted DualQuaternion
        :rtype: DualQuaternion
        """
        p = self.p - other.p
        d = self.d - other.d
        return DualQuaternion.from_two_quaternions(p, d)

    def __mul__(self, other) -> "DualQuaternion":
        """
        Multiplication of two DualQuaternions

        :param DualQuaternion, int, float other: other DualQuaternion

        :return: multiplied DualQuaternion
        :rtype: DualQuaternion
        """
        if isinstance(other, (int, float)):
            return DualQuaternion(self.array() * other)
        else:
            p = self.p * other.p
            d = (self.d * other.p) + (self.p * other.d)
            return DualQuaternion.from_two_quaternions(p, d)

    def __rmul__(self, other) -> "DualQuaternion":
        """Handle when the dual quaternion is on the right side of the multiplication"""
        return self.__mul__(other)

    def __truediv__(self, other) -> "DualQuaternion":
        """
        Division of two DualQuaternions

        :param DualQuaternion, int, float other: other DualQuaternion

        :return: divided DualQuaternion
        :rtype: DualQuaternion
        """
        if isinstance(other, (int, float)):
            return DualQuaternion(self.array() / other)
        else:
            return self * other.inv()

    def __neg__(self) -> "DualQuaternion":
        """
        Negation of the DualQuaternion

        :return: negated DualQuaternion
        :rtype: DualQuaternion
        """
        return DualQuaternion(-1 * self.array())

    def real(self) -> np.ndarray:
        """
        Real part of the DualQuaternion, list of first and fifth element

        :return: real part of the DualQuaternion
        :rtype: np.ndarray
        """
        return np.array([self.array()[0], self.array()[4]])

    def imag(self) -> np.ndarray:
        """
        Imaginary part of the DualQuaternion

        List of 6 elements - ijk elements of the primal and dual part. In case of a
        line, these might be Plucker coordinates.

        :return: imaginary part of the DualQuaternion
        :rtype: np.ndarray
        """
        return np.array([self.array()[1], self.array()[2], self.array()[3],
                         self.array()[5], self.array()[6], self.array()[7]])

    def back_projection(self) -> "DualQuaternion":
        """
        Returns the projection of dual quaternion onto Study quadric.

        The back projection, or also known as fiber projection, is a method to project
        a dual quaternion onto the Study quadric, i.e. to obtain a proper rigid body
        displacement that it represents.

        :return: DualQuaternion
        :rtype: DualQuaternion
        """
        primal = self.p
        dual = self.d

        primal_2norm = 2 * primal.norm()
        new_primal = Quaternion(np.array([primal_2norm, 0, 0, 0]))
        new_dual = -1 * (primal * dual.conjugate() - dual * primal.conjugate())

        dq = (DualQuaternion.from_two_quaternions(new_primal, new_dual)
              * DualQuaternion.from_two_quaternions(primal, Quaternion(np.zeros(4)))
              ) / 2

        return dq

    def array(self) -> np.ndarray:
        """
        DualQuaternion to numpy array (8-vector of study parameters)

        :return: DualQuaternion as numpy array
        :rtype: np.ndarray
        """
        return np.concatenate((self.p.array(), self.d.array()))

    def conjugate(self) -> "DualQuaternion":
        """
        Dual Quaternion conjugate

        :return: conjugated DualQuaternion
        :rtype: DualQuaternion
        """
        return DualQuaternion.from_two_quaternions(
            self.p.conjugate(), self.d.conjugate())

    def eps_conjugate(self) -> "DualQuaternion":
        """
        Dual Quaternion epsilon conjugate

        :return: epsilon-conjugated DualQuaternion
        :rtype: DualQuaternion
        """
        dual_part_eps_c = -1 * self.d.array()
        return DualQuaternion(np.concatenate((self.p.array(), dual_part_eps_c)))

    def norm(self) -> "DualQuaternion":
        """
        Dual Quaternion norm as dual number (8-vector of study parameters), primal norm
        is in the first element, dual norm is in the fifth element

        :return: norm of the DualQuaternion
        :rtype: DualQuaternion
        """
        n = self.p.norm()
        eps_n = 2 * (
            self.p[0] * self.d[0]
            + self.p[1] * self.d[1]
            + self.p[2] * self.d[2]
            + self.p[3] * self.d[3]
        )
        return DualQuaternion(np.array([n, 0, 0, 0, eps_n, 0, 0, 0]))

    def inv(self):
        """
        Inverse of the DualQuaternion

        :return: inverse of the DualQuaternion
        :rtype: DualQuaternion
        """
        p = self.p.inv()
        d = -1 * p * self.d * p
        return DualQuaternion.from_two_quaternions(p, d)

    def is_on_study_quadric(self) -> bool:
        """
        Check if the DualQuaternion is on the study quadric

        :return: True if the DualQuaternion is on the study quadric, False otherwise
        :rtype: bool
        """
        study_condition = (self.p[0] * self.d[0] + self.p[1] * self.d[1]
                           + self.p[2] * self.d[2] + self.p[3] * self.d[3])
        return np.isclose(study_condition, 0)

    def dq2matrix(self):
        """
        Dual Quaternion to SE(3) transformation matrix

        :return: 4x4 transformation matrix
        :rtype: np.ndarray
        """
        p0 = self[0]
        p1 = self[1]
        p2 = self[2]
        p3 = self[3]
        d0 = self[4]
        d1 = self[5]
        d2 = self[6]
        d3 = self[7]

        # mapping
        r11 = p0**2 + p1**2 - p2**2 - p3**2
        r22 = p0**2 - p1**2 + p2**2 - p3**2
        r33 = p0**2 - p1**2 - p2**2 + p3**2
        r44 = p0**2 + p1**2 + p2**2 + p3**2

        r12 = 2 * (p1 * p2 - p0 * p3)
        r13 = 2 * (p1 * p3 + p0 * p2)
        r21 = 2 * (p1 * p2 + p0 * p3)
        r23 = 2 * (p2 * p3 - p0 * p1)
        r31 = 2 * (p1 * p3 - p0 * p2)
        r32 = 2 * (p2 * p3 + p0 * p1)

        r14 = 2 * (-p0 * d1 + p1 * d0 - p2 * d3 + p3 * d2)
        r24 = 2 * (-p0 * d2 + p1 * d3 + p2 * d0 - p3 * d1)
        r34 = 2 * (-p0 * d3 - p1 * d2 + p2 * d1 + p3 * d0)

        tr = np.array(
            [
                [r44, 0, 0, 0],
                [r14, r11, r12, r13],
                [r24, r21, r22, r23],
                [r34, r31, r32, r33],
            ]
        )

        # Normalization
        output_matrix = tr / tr[0, 0]
        return output_matrix

    def dq2point_via_matrix(self) -> np.ndarray:
        """
        Dual Quaternion to point via SE(3) transformation matrix

        :return: array of 3-coordinates of point
        :rtype: np.ndarray
        """
        mat = self.dq2matrix()
        return mat[1:4, 0]

    def dq2point(self) -> np.ndarray:
        """
        Dual Quaternion directly to point

        :return: array of 3-coordinates of point
        :rtype: np.ndarray
        """
        dq = self.array() / self.array()[0]
        return dq[5:8]

    def dq2point_homogeneous(self) -> np.ndarray:
        """
        Dual Quaternion directly to point

        :return: array of 3-coordinates of point
        :rtype: np.ndarray
        """
        dq = self.array()
        return np.array([dq[0], dq[5], dq[6], dq[7]])

    def dq2line_vectors(self) -> tuple:
        """
        Dual Quaternion directly to line coordinates

        If the DQ is a sympy Expression, it is not converted to float and not normalized

        :return: tuple of 2 numpy arrays, 3-vector coordinates each
        :rtype: tuple

        :raises ValueError: if the dual quaternion has more than one indeterminate

        :warn: if the dual quaternion has NOT zeros in 1st and 5th element, it is not
            a line
        """
        dq = self.array()
        if any(isinstance(x, Expr) for x in dq):
            try:
                dq = np.asarray(dq, dtype="float64")
            except Exception:
                pass

        if any(isinstance(x, Expr) for x in dq):
            dq0_simple = simplify(dq[0])
            dq4_simple = simplify(dq[4])

            if len((dq0_simple + dq4_simple).free_symbols) > 1:
                raise ValueError("dq2line_vectors method error: the dual quaternion "
                                 "has more than one indeterminate.")

            if not (dq0_simple == 0 or dq4_simple == 0):
                coeffs = []
                for t in dq0_simple.free_symbols:
                    coeffs += Poly(dq0_simple, t).all_coeffs()
                for t in dq4_simple.free_symbols:
                    coeffs += Poly(dq4_simple, t).all_coeffs()

                coeffs = np.array(coeffs, dtype="float64")
                all_close_to_zero = all(np.isclose(coeff, 0) for coeff in coeffs)

                if not all_close_to_zero:  # warn that the DQ is not a line
                    warn("dq2line_vectors method warning: the dual quaternion has NOT "
                         "zeros in 1st and 5th element, it is not a line.")

            direction = dq[1:4]
            moment = -1 * dq[5:8]
            # TODO normalize the direction and moment when sympy expressions are used

        else:
            k = dq[0] ** 2 - dq[1] ** 2 - dq[2] ** 2 - dq[3] ** 2  # differs from Study
            f = k - dq[0] ** 2
            g = dq[0] * dq[4]

            dir = f * dq[1:4]
            mom = np.array([g * dq[1] - f * dq[5],
                            g * dq[2] - f * dq[6],
                            g * dq[3] - f * dq[7]])

            moment = -1 * mom / np.linalg.norm(dir)
            direction = -1 * dir / np.linalg.norm(dir)

        return direction, moment

    def dq2screw(self) -> np.ndarray:
        """
        Dual Quaternion directly to screw coordinates

        :return: array of 6-coordinates of screw
        :rtype: np.ndarray
        """
        direction, moment = self.dq2line_vectors()
        return np.concatenate((direction, moment))

    def dq2point_via_line(self) -> np.ndarray:
        """
        Dual Quaternion to point via line coordinates

        :return: array of 3-coordinates of point
        :rtype: np.ndarray
        """
        direction, moment = self.dq2line_vectors()
        return np.cross(direction, moment)

    def act(
        self,
        affected_object: Union["DualQuaternion", "NormalizedLine", "PointHomogeneous"],
    ) -> Union["NormalizedLine", "PointHomogeneous"]:
        """
        Act on a line or point with the DualQuaternion

        The action of a DualQuaternion is a half-turn about its axis. If the
        acted_object is a DualQuaternion (rotation axis DQ), it is converted to
        NormalizedLine and then the action is performed.

        :param DualQuaternion, NormalizedLine, or PointHomogeneous affected_object:
            object to act on (line or point)

        :return: line or point
        :rtype: NormalizedLine, PointHomogeneous

        :examples:

        .. testcode::

            # Act on a line with a dual quaternion

            from rational_linkages import DualQuaternion, NormalizedLine


            dq = DualQuaternion([1, 0, 0, 1, 0, 3, 2, -1])
            line = NormalizedLine.from_direction_and_point([0, 0, 1], [0, -2, 0])

            line_after_half_turn = dq.act(line)
        """
        from .DualQuaternionAction import DualQuaternionAction

        action = DualQuaternionAction()
        return action.act(self, affected_object)



from typing import Optional, Sequence, Union
from warnings import warn

import numpy as np
from sympy import Expr

# Forward declarations for class names
DualQuaternion = "DualQuaternion"
PointHomogeneous = "PointHomogeneous"


class NormalizedLine:
    """
    Class representing a Normalized Line in Dual Quaternion space.

    The NormalizedLine is defined using Plucker coordinates, representing a Unit Screw
    axis.

    :param np.ndarray, list[float] unit_screw: Plucker coordinates representing the
        Unit Screw axis. If None, the line is in the origin along the Z axis.

    :ivar np.ndarray direction: Direction vector of the normalized line.
    :ivar np.ndarray moment: Moment vector of the normalized line.
    :ivar np.ndarray screw: Concatenation of the direction and moment vectors.
    :ivar np.ndarray as_dq_array: Conversion of the NormalizedLine to a Dual Quaternion
        array.

    :examples:

    .. testcode::

        # Creating a NormalizedLine from a unit screw axis

        from rational_linkages import NormalizedLine
        line = NormalizedLine([1, 0, 0, 0, -2, 1])

    .. testcode::

        # Creating a default NormalizedLine at the origin along the Z axis

        from rational_linkages import NormalizedLine
        line = NormalizedLine()

    .. testcode::

        # Creating a NormalizedLine from two points

        from rational_linkages import NormalizedLine
        from rational_linkages import PointHomogeneous
        point1 = PointHomogeneous([1, 1, 1, 1])
        point2 = PointHomogeneous([1, 3, 1, 1])
        line = NormalizedLine.from_two_points(point1, point2)

    .. testcode::

        # Creating a NormalizedLine from a direction and a point

        from rational_linkages import NormalizedLine
        line = NormalizedLine.from_direction_and_point([1, 0, 0], [1, 1, 1])

    .. testcode::

        # Creating a NormalizedLine from a direction and a moment

        from rational_linkages import NormalizedLine
        line = NormalizedLine.from_direction_and_moment([1, 0, 0], [0, 1, -1])

    .. testcode::

        # Creating a NormalizedLine from a DualQuaternion

        from rational_linkages import NormalizedLine
        from rational_linkages import DualQuaternion
        dq = DualQuaternion([0, 0, 0, 1, 0, 0, 0, 0])
        line = NormalizedLine.from_dual_quaternion(dq)
    """

    def __init__(self,
                 unit_screw: Optional[Sequence[Union[float, np.ndarray]]] = None):
        """
        Normalized line class in Dual Quaternion space

        Given by Plucker coordinates, representing a Unit Screw axis

        :param np.ndarray, list[float] unit_screw: Plucker coordinates
            representing the Unit Screw axis
        """
        if unit_screw is None:
            # in origin along Z axis
            unit_screw = np.array([0, 0, 1, 0, 0, 0])
        elif any(isinstance(element, Expr) for element in unit_screw):
            # sympy object, try to convert it to numpy float
            try:
                unit_screw = np.asarray(unit_screw, dtype='float64')
            except Exception:
                self.direction = np.asarray(unit_screw[0:3])
                self.moment = np.asarray(unit_screw[3:6])

        if not any(isinstance(element, Expr) for element in unit_screw):
            direction = np.asarray(unit_screw[0:3])
            moment = np.asarray(unit_screw[3:6])
            # Check if the direction vector is normalized
            if round(np.linalg.norm(direction), 6) == 1.0:
                self.direction = direction
                self.moment = moment
            elif np.abs(np.linalg.norm(direction)) > 1e-10:
                self.direction = direction / np.linalg.norm(direction)
                self.moment = moment / np.linalg.norm(direction)
            else:
                warn("Direction vector has zero norm!")
                self.direction = np.asarray(direction)
                self.moment = np.asarray(moment)

        self.screw = np.concatenate((self.direction, self.moment))

    def __repr__(self):
        line = np.array2string(self.screw,
                               precision=10,
                               suppress_small=True,
                               separator=", ",
                               max_line_width=100000)
        return f"{line}"

    @classmethod
    def from_two_points(cls,
                        pt0: Union['PointHomogeneous', list[float, float, float]],
                        pt1: Union['PointHomogeneous', list[float, float, float]]
                        ) -> "NormalizedLine":
        """
        Construct NormalizedLine from two points

        :param np.ndarray, list[float] pt0: PointHomogeneous or list or np.array of
            shape (3,)
        :param np.ndarray, list[float] pt1: PointHomogeneous or list or np.array of
            shape (3,)

        :return: NormalizedLine
        :rtype: NormalizedLine
        """
        from .PointHomogeneous import PointHomogeneous

        if isinstance(pt0, PointHomogeneous) and isinstance(pt1, PointHomogeneous):
            pt0 = pt0.normalized_in_3d()
            pt1 = pt1.normalized_in_3d()
        else:
            pt0 = np.asarray(pt0)
            pt1 = np.asarray(pt1)

        if np.allclose(pt0, pt1, rtol=1e-7):
            raise ValueError("Points are the same!")

        direction = np.asarray(pt1 - pt0)
        moment = np.cross(-1 * direction, np.asarray(pt0))
        return cls(np.concatenate((direction, moment)))

    @classmethod
    def from_direction_and_point(
        cls, direction: [float, float, float], point: [float, float, float]
    ) -> "NormalizedLine":
        """
        Construct NormalizedLine from direction and point

        :param np.ndarray, list[float] direction: list or np.array of shape (3,)
        :param np.ndarray, list[float] point: list or np.array of shape (3,)

        :return: NormalizedLine
        :rtype: NormalizedLine
        """
        direction = np.asarray(direction)
        point = np.asarray(point)
        moment = np.cross(-1 * direction, point)
        return cls(np.concatenate((direction, moment)))

    @classmethod
    def from_direction_and_moment(
        cls, direction: [float, float, float], moment: [float, float, float]
    ) -> "NormalizedLine":
        """
        Construct NormalizedLine from direction and moment

        :param np.ndarray, list[float] direction: list or np.array of shape (3,)
        :param np.ndarray, list[float] moment: list or np.array of shape (3,)

        :return: NormalizedLine
        :rtype: NormalizedLine
        """
        direction = np.asarray(direction)
        moment = np.asarray(moment)
        return cls(np.concatenate((direction, moment)))

    @classmethod
    def from_dual_quaternion(cls, dq: "DualQuaternion") -> "NormalizedLine":
        """
        Construct NormalizedLine from DualQuaternion

        :param DualQuaternion dq: DualQuaternion

        :return: NormalizedLine from DualQuaternion
        :rtype: NormalizedLine
        """
        return cls(dq.dq2screw())

    def line2dq_array(self) -> np.ndarray:
        """
        Embed NormalizedLine to array of floats representing the unit screw in
        the form of Dual Quaternion

        :return: np.array of shape (8,)
        :rtype: np.ndarray
        """
        return np.array(
            [
                0,
                self.direction[0],
                self.direction[1],
                self.direction[2],
                0,
                -1 * self.moment[0],
                -1 * self.moment[1],
                -1 * self.moment[2],
            ]
        )

    def point_on_line(self, t: float = 0.0) -> np.array:
        """
        Get principal point on axis

        :param float t: t-point parameter

        :return: numpy array 3-vector point coordinates
        :rtype: np.ndarray
        """
        principal_point = np.cross(self.direction, self.moment)
        return principal_point + (t * self.direction)

    def point_homogeneous(self) -> np.array:
        """
        Get a homogeneous coordinate of a point on Plucker line; choose point with
        the highest value in the first column

        :return: numpy array 4-vector point coordinates
        :rtype: np.ndarray
        """
        pt_quadric = np.array(np.zeros((3, 4)))
        # pt_quadric = [0, self.direction[0], self.direction[1], self.direction[2]]
        pt_quadric[0, :] = [-self.direction[0], 0, self.moment[2], -self.moment[1]]
        pt_quadric[1, :] = [-self.direction[1], -self.moment[2], 0, self.moment[0]]
        pt_quadric[2, :] = [-self.direction[2], self.moment[1], -self.moment[0], 0]

        abs_points_1st_column = abs(pt_quadric[:, 0])
        max_index = abs_points_1st_column.argmax()
        # return pt_quadric[2, :]
        return pt_quadric[max_index, :]

    def get_point_param(self, point: Union[np.ndarray, list[float, float, float]]) -> (
            np.ndarray):
        """
        Get a parameter for a given point that lies the line

        :param np.ndarray, list[float] point: np.array of shape (3,)

        :return: parameter for the point on the joint
        :rtype: np.ndarray
        """
        # vector between given point and principal point
        point = np.asarray(point)
        vec = point - self.point_on_line()

        # avoid situation if the direction vector is parallel to one of the origin axes
        for i in range(3):
            if self.direction[i] != 0.0:
                return vec[i] / self.direction[i]

        raise ValueError("Direction vector is zero!")

    def common_perpendicular_to_other_line(self, other) -> tuple:
        """
        Get the common perpendicular to another Plucker line (two intersection points
        and distance).

        :param NormalizedLine other: other normalized line

        :return: points, distance, cos_angle
        :rtype: tuple
        """
        # Initialize arrays to store the intersection points
        points = [np.zeros(3), np.zeros(3)]

        # Calculate the cross product of the direction vectors
        cross_product = np.cross(self.direction, other.direction)
        cross_product_norm = np.linalg.norm(cross_product)

        # if lines are not parallel
        if not np.isclose(cross_product_norm, 0.0, atol=1e-5):
            # Calculate the first intersection point
            numerator1 = np.cross(
                -self.moment, np.cross(other.direction, cross_product)
            ) + np.dot(self.direction, np.dot(other.moment, cross_product))
            points[0] = numerator1 / (cross_product_norm**2)

            # Calculate the second intersection point
            numerator2 = np.cross(
                other.moment, np.cross(self.direction, cross_product)
            ) - np.dot(other.direction, np.dot(self.moment, cross_product))
            points[1] = numerator2 / (cross_product_norm**2)

            # Calculate the distance and cosine of the angle between the lines
            distance = np.linalg.norm(points[0] - points[1])
            cos_angle = np.dot(self.direction, other.direction) / (
                np.linalg.norm(self.direction) * np.linalg.norm(other.direction)
            )
        else:
            # Lines are parallel, use alternative approach
            points[0] = np.cross(self.direction, self.moment)
            points[1] = np.cross(other.direction, other.moment)

            # # legacy code, not working properly
            # vec = np.cross(self.direction, self.moment - other.moment) / (
            #     np.linalg.norm(self.direction) ** 2
            # )
            # vec = np.array(vec, dtype="float64")
            # distance = np.linalg.norm(vec)

            distance = np.linalg.norm(points[0] - points[1])
            cos_angle = 1.0

        return points, distance, cos_angle

    def contains_point(self, point: Union['PointHomogeneous', np.ndarray, list[float, float, float]]) -> bool:
        """
        Check if the line contains given point

        The method basically creates a new line moment from the given point and
        the direction of the line. If they create the same line (moment), the point is
        on the line.

        :param PointHomogeneous, np.ndarray, list[float] point: point of shape (3,)

        :return: True if the point is on the line, False otherwise
        :rtype: bool
        """
        from .PointHomogeneous import PointHomogeneous

        if isinstance(point, PointHomogeneous):
            point = point.normalized_in_3d()
        else:
            point = np.asarray(point)

        return np.allclose(np.cross(point, self.direction), self.moment)

    def get_plot_data(self, interval: tuple) -> np.ndarray:
        """
        Get data for plotting the line in 3D

        :param tuple interval: interval of the parameter t (start, end of the line)

        :return: starting point and vector direction of shape (6, 1)
        :rtype: np.ndarray
        """
        # points on the line
        p0 = self.point_on_line(interval[0])
        p1 = self.point_on_line(interval[1])
        # vector between points
        vec = p1 - p0

        return np.concatenate((p0, vec))

    def evaluate(self, t_param: float) -> 'NormalizedLine':
        """
        Evaluate the line at the given parameter

        :param float t_param: parameter

        :return: evaluated line with float elements
        :rtype: NormalizedLine
        """
        from sympy import Expr, Symbol, Number

        t = Symbol("t")

        line_expr = [Expr(coord) if not isinstance(coord, Number) else coord
                     for coord in self.screw]
        line = [coord.subs(t, t_param).evalf().args[0]
                if not isinstance(coord, Number) else coord
                for coord in line_expr]
        return NormalizedLine(np.asarray(line, dtype="float64"))






from functools import wraps
from itertools import cycle

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, TextBox

from .DualQuaternion import DualQuaternion
from .MotionFactorization import MotionFactorization
from .NormalizedLine import NormalizedLine
from .PointHomogeneous import PointHomogeneous
from .RationalBezier import RationalBezier
from .RationalCurve import RationalCurve
from .RationalMechanism import RationalMechanism
from .TransfMatrix import TransfMatrix


class Plotter:
    def __init__(self,
                 interactive: bool = False,
                 jupyter_notebook: bool = False,
                 interval=(-1, 1),
                 steps=50,
                 arrows_length: float = 1.0,
                 joint_range_lim: float = 1.0):
        """
        Initialize the plotter

        :param interactive: activate interactive mode
        :param jupyter_notebook: activate jupyter notebook mode
        :param interval: interval for plotting
        :param steps: number of steps for plotting
        :param arrows_length: length of quiver arrows for poses and frames
        :param joint_range_lim: limit for joint sliders, will be +/- value
        """
        # use Qt5Agg backend for interactive plotting
        if interactive and not jupyter_notebook:
            matplotlib.use("Qt5Agg")

        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(projection="3d")

        self.ax.set_xlabel("X-axis")
        self.ax.set_ylabel("Y-axis")
        self.ax.set_zlabel("Z-axis")
        self.ax.set_aspect("equal")

        # Initialize min/max variables
        self.min_x, self.max_x = float('inf'), float('-inf')
        self.min_y, self.max_y = float('inf'), float('-inf')
        self.min_z, self.max_z = float('inf'), float('-inf')

        if interactive:
            plt.subplots_adjust(
                top=1.0,
                bottom=0.16,
                left=0.32,
                right=0.935,
                hspace=0.2,
                wspace=0.2
            )
        else:
            plt.subplots_adjust(
                top=1.0,
                bottom=0.1,
                left=0.0,
                right=1.0,
                hspace=0.2,
                wspace=0.2
            )

        self.t_space = np.linspace(interval[0], interval[1], steps)
        self.steps = steps
        self.interactive = interactive
        self.jupyter_notebook = jupyter_notebook
        self.j_sliders_limit = joint_range_lim

        # length of quiver arrows for poses and frames
        self.arrows_length = arrows_length

        self.plotted = {}

    def plot(self, object_to_plot, **kwargs):
        """
        Plot the object

        :param object_to_plot: NormalizedLine, PointHomogeneous, RationalMechanism,
            MotionFactorization, DualQuaternion, TransfMatrix, RationalCurve
            or RationalBezier
        :param kwargs: plotting options following matplotlib standards and syntax

        :return: matplotlib axis
        """
        type_to_plot = self.analyze_object(object_to_plot)

        match type_to_plot:
            case "is_line":
                self._plot_line(object_to_plot, **kwargs)
            case "is_point":
                self._plot_point(object_to_plot, **kwargs)
            case "is_motion_factorization":
                self._plot_motion_factorization(object_to_plot, **kwargs)
            case "is_dq":
                self._plot_dual_quaternion(object_to_plot, **kwargs)
            case "is_transf_matrix":
                self._plot_transf_matrix(object_to_plot, **kwargs)
            case "is_rational_curve":
                self._plot_rational_curve(object_to_plot, **kwargs)
            case "is_rational_bezier":
                self._plot_rational_bezier(object_to_plot, **kwargs)
            case "is_rational_mechanism":
                self._plot_rational_mechanism(object_to_plot, **kwargs)
            case "is_interactive":
                self._plot_interactive(object_to_plot, **kwargs)

    def analyze_object(self, object_to_plot):
        """
        Analyze the object to plot

        :param object_to_plot: NormalizedLine, PointHomogeneous, RationalMechanism,
            MotionFactorization, DualQuaternion, TransfMatrix, RationalCurve
            or RationalBezier

        :return: str - 'is_line', 'is_point', 'is_motion_factorization', 'is_dq' or
            'is_rational_mechanism'
        """
        if isinstance(object_to_plot, RationalMechanism) and not self.interactive:
            return "is_rational_mechanism"
        elif isinstance(object_to_plot, RationalMechanism) and self.interactive:
            return "is_interactive"
        elif isinstance(object_to_plot, MotionFactorization) and not self.interactive:
            return "is_motion_factorization"
        elif isinstance(object_to_plot, NormalizedLine):
            return "is_line"
        elif isinstance(object_to_plot, PointHomogeneous):
            return "is_point"
        elif isinstance(object_to_plot, RationalBezier):
            return "is_rational_bezier"
        elif isinstance(object_to_plot, RationalCurve):
            return "is_rational_curve"
        elif isinstance(object_to_plot, DualQuaternion):
            return "is_dq"
        elif isinstance(object_to_plot, TransfMatrix):
            return "is_transf_matrix"
        else:
            raise TypeError(
                "Other types than NormalizedLine, PointHomogeneous, RationalMechanism, "
                "MotionFactorization or DualQuaternion not yet implemented"
            )

    @staticmethod
    def _plotting_decorator(func):
        """
        Decorator for plotting functions

        :param func: plotting function
        :return: decorated plotting function
        """
        @wraps(func)
        def _wrapper(self, *args, **kwargs):
            # use the plotting function
            func(self, *args, **kwargs)

            # decorate the plot - set aspect ratio and update legend
            self.ax.set_aspect("equal")
            if not self.jupyter_notebook:
                self.ax.legend()
        return _wrapper

    @_plotting_decorator
    def _plot_line(self, line: NormalizedLine, **kwargs):
        """
        Plot a line

        :param line: NormalizedLine
        :param kwargs: matplotlib options
        """
        if 'interval' in kwargs:
            interval = kwargs['interval']
            kwargs.pop('interval')
        else:
            interval = (-1, 1)

        line = line.get_plot_data(interval)

        if 'label' not in kwargs:
            kwargs['label'] = "line"
        else:
            mid_of_line = (line[:3] + line[3:]/2)
            self.ax.text(*mid_of_line, ' ' + kwargs['label'])

        self.ax.quiver(*line, **kwargs)

    @_plotting_decorator
    def _plot_point(self, point: PointHomogeneous, **kwargs):
        """
        Plot a point

        :param point: PointHomogeneous
        :param kwargs: matplotlib options
        """
        point = point.get_plot_data()

        if 'label' not in kwargs:
            kwargs['label'] = "point"
        else:
            self.ax.text(*point, ' ' + kwargs['label'])

        self.ax.scatter(*point, **kwargs)

    @_plotting_decorator
    def _plot_dual_quaternion(self, dq: DualQuaternion, **kwargs):
        """
        Plot a dual quaternion as a transformation

        :param dq: DualQuaternion
        :param kwargs: not used
        """
        matrix = TransfMatrix(dq.dq2matrix())
        self._plot_transf_matrix(matrix, **kwargs)

    @_plotting_decorator
    def _plot_transf_matrix(self, matrix: TransfMatrix, **kwargs):
        """
        Plot a transformation matrix

        :param transf_matrix: TransfMatrix
        :param kwargs: not used
        """
        x_vec, y_vec, z_vec = matrix.get_plot_data()

        if 'label' not in kwargs:
            kwargs['label'] = 'Tf'
        else:
            self.ax.text(*matrix.t, ' ' + kwargs['label'])

        self.ax.quiver(*x_vec, color="red", length=self.arrows_length)
        self.ax.quiver(*y_vec, color="green", length=self.arrows_length)
        self.ax.quiver(*z_vec, color="blue", length=self.arrows_length)

    @_plotting_decorator
    def _plot_rational_curve(self, curve: RationalCurve, **kwargs):
        """
        Plot a rational curve

        :param RationalCurve curve: RationalCurve
        :param kwargs: interval and matplotlib options
        """
        if 'interval' in kwargs:
            interval = kwargs['interval']
            kwargs.pop('interval')
        else:
            interval = (0, 1)

        x, y, z = curve.get_plot_data(interval, self.steps)

        if 'label' not in kwargs:
            kwargs['label'] = 'curve'

        self.ax.plot(x, y, z, **kwargs)

    @_plotting_decorator
    def _plot_rational_bezier(self, bezier: RationalBezier, **kwargs):
        """
        Plot a rational Bezier curve

        :param bezier: RationalBezier
        :param kwargs: interval and matplotlib options
        """
        if 'interval' in kwargs:
            interval = kwargs['interval']
            kwargs.pop('interval')
        else:
            interval = (0, 1)

        x, y, z, x_cp, y_cp, z_cp = bezier.get_plot_data(interval, self.steps)

        if 'label' not in kwargs:
            kwargs['label'] = "bezier curve"

        self.ax.plot(x, y, z, **kwargs)
        self.ax.plot(x_cp, y_cp, z_cp, "ro:")

    @_plotting_decorator
    def _plot_motion_factorization(self, factorization: MotionFactorization, **kwargs):
        """
        Plot a motion factorization

        :param factorization: MotionFactorization
        :param kwargs: t-curve parameter of driving joint axis and matplotlib options
        """
        if 't' in kwargs:
            t = kwargs['t']
            kwargs.pop('t')
        else:
            t = 0

        points = factorization.direct_kinematics(t)
        x, y, z = zip(*points)

        if 'label' not in kwargs:
            kwargs['label'] = "factorization"

        self.ax.plot(x, y, z, **kwargs)

    @_plotting_decorator
    def _plot_rational_mechanism(self, mechanism: RationalMechanism, **kwargs):
        """
        Plot a mechanism

        :param mechanism: RationalMechanism
        :param kwargs: t-curve parameter of driving joint axis and matplotlib options
        """
        self.plotted['mechanism'] = mechanism
        if 't' in kwargs:
            t = kwargs['t']
            kwargs.pop('t')
        else:
            t = 0

        # plot factorizations
        for factorization in mechanism.factorizations:
            self._plot_motion_factorization(factorization, t=t, **kwargs)

        # plot end effector triangle
        pts0 = mechanism.factorizations[0].direct_kinematics_of_tool_with_link(t, mechanism.tool_frame.dq2point_via_matrix())
        pts1 = mechanism.factorizations[1].direct_kinematics_of_tool_with_link(t, mechanism.tool_frame.dq2point_via_matrix())[::-1]
        ee_points = np.concatenate((pts0, pts1))

        if 'label' not in kwargs:
            kwargs['label'] = "end effector"

        x, y, z = zip(*ee_points)
        self.ax.plot(x, y, z, **kwargs)

        self._plot_tool_path(mechanism, **kwargs)

    def _plot_tool_path(self, mechanism: RationalMechanism, **kwargs):
        # plot end effector path
        t_lin = np.linspace(0, 2 * np.pi, self.steps)
        t = [mechanism.factorizations[0].joint_angle_to_t_param(t_lin[i])
             for i in range(self.steps)]

        ee_points = [mechanism.factorizations[0].direct_kinematics_of_tool(
            t[i], mechanism.tool_frame.dq2point_via_matrix()) for i in range(self.steps)]

        kwargs['label'] = "tool path"

        x, y, z = zip(*ee_points)
        self.ax.plot(x, y, z, **kwargs)

    @_plotting_decorator
    def _plot_interactive(self,
                          mechanism: RationalMechanism,
                          show_tool: bool = True,
                          **kwargs):
        """
        Plot a mechanism in interactive mode

        :param RationalMechanism mechanism: RationalMechanism
        :param bool show_tool: show tool linkage and frame
        :param kwargs: matplotlib options
        """
        self.plotted['mechanism'] = mechanism
        self.show_tool = show_tool

        # plot the curve (tool path)
        self._plot_tool_path(mechanism, **kwargs)

        # append first slider that is the driving joint angle slider
        self.move_slider = self._init_slider()
        # set a text box that can be used to set the angle manually
        self.text_box_angle = TextBox(self.fig.add_axes([0.3, 0.055, 0.15, 0.05]),
                                      "Set angle [rad]: ", textalignment="right")
        # set a text box that can be used to set the t param manually
        self.text_box_param = TextBox(self.fig.add_axes([0.3, 0.11, 0.15, 0.05]),
                                      "Set param t [-]: ", textalignment="right")
        # text box to save files
        self.text_box_save = TextBox(self.fig.add_axes([0.3, 0.165, 0.15, 0.05]),
                                     "Save with filename: ", textalignment="right")

        # vertical sliders to control physical linkage position (connecting points)
        self.joint_sliders = []
        for i in range(mechanism.num_joints):
            slider0, slider1 = self._init_slider(idx=i,
                                                 j_sliders=self.joint_sliders,
                                                 slider_limit=self.j_sliders_limit)
            self.joint_sliders.append(slider0)
            self.joint_sliders.append(slider1)

        # set joint parameters to home configuration
        for i in range(mechanism.factorizations[0].number_of_factors):
            self.joint_sliders[2 * i].set_val(
                mechanism.factorizations[0].linkage[i].points_params[0])
            self.joint_sliders[1 + 2 * i].set_val(
                mechanism.factorizations[0].linkage[i].points_params[1])

        for i in range(mechanism.factorizations[1].number_of_factors):
            self.joint_sliders[
                2 * mechanism.factorizations[0].number_of_factors + 2 * i].set_val(
                mechanism.factorizations[1].linkage[i].points_params[0])
            self.joint_sliders[2 * mechanism.factorizations[
                0].number_of_factors + 1 + 2 * i].set_val(
                mechanism.factorizations[1].linkage[i].points_params[1])

        # initialize the linkages plot
        linestyles = cycle(['solid', 'dashdot'])
        self.lines = []
        for i in range(mechanism.num_joints * 2):
            # alter between solid (links) and dashdot (joints)
            linestyle = next(linestyles)
            line, = self.ax.plot([], [], [], linestyle=linestyle,
                                 color='black', marker='.')
            self.lines.append(line)

        if self.show_tool:
            # initialize the tool point interactive plot
            self.tool_plot, = self.ax.plot([], [], [], color="purple",
                                           linestyle="dashed", label="tool connection")
            # initialize the tool frame
            self.pose_frame = [self.ax.quiver([], [], [], [], [], [], color="red",
                                              length=self.arrows_length),
                               self.ax.quiver([], [], [], [], [], [], color="green",
                                              length=self.arrows_length),
                               self.ax.quiver([], [], [], [], [], [], color="blue",
                                              length=self.arrows_length)]

        def submit_angle(text):
            """Event handler for the text box"""
            val = float(text)
            val = val % (2 * np.pi)
            self.move_slider.set_val(val)

        def submit_parameter(text):
            """Event handler for the text box"""
            val = float(text)
            self.plot_slider_update(val, t_param=val)
            self.move_slider.set_val(mechanism.factorizations[0].t_param_to_joint_angle(val))

        def submit_save(text):
            """Event handler for the text box"""
            val = text
            mechanism.save(filename=val)

        # connect the slider and text box to the event handlers
        self.move_slider.on_changed(self.plot_slider_update)
        self.text_box_angle.on_submit(submit_angle)
        self.text_box_param.on_submit(submit_parameter)
        self.text_box_save.on_submit(submit_save)

        # joint physical placement sliders
        for i in range(4 * mechanism.factorizations[0].number_of_factors):
            self.joint_sliders[i].on_changed(self.plot_connecting_points_update)

        # initialize the plot in home configuration
        self.move_slider.set_val(0.0)

    @staticmethod
    def _init_slider(idx: int = None, j_sliders=None, slider_limit: float = 1.0):
        """
        Initialize the slider for interactive plotting

        :param int idx: index of the slider, first one is added automatically as joint
            angle slider
        :param list j_sliders: list of joint sliders
        :param float slider_limit: limit for joint sliders, will be +/- value

        :return: matplotlib slider
        """
        if idx is None:  # driving joint angle slider
            slider = Slider(
                ax=plt.axes([0.3, 0.01, 0.5, 0.05]),
                label="Joint angle [rad]: ",
                valmin=0.0,
                valmax=np.pi * 2,
                valinit=0.0,
                valstep=0.01,
            )
            return slider

        else:  # joint connection points sliders
            i = int(len(j_sliders) / 2)
            slider0 = Slider(
                ax=plt.axes([0.03 + i * 0.04, 0.25, 0.0225, 0.63]),
                #label="j{}.0".format(i),
                label="j{}".format(i),
                valmin=-slider_limit,
                valmax=slider_limit,
                valinit=0.0,
                orientation="vertical",
            )
            slider1 = Slider(
                ax=plt.axes([0.045 + i * 0.04, 0.25, 0.0225, 0.63]),
                #label="j{}.1".format(i),
                label="",
                valmin=-slider_limit,
                valmax=slider_limit,
                valinit=0.0,
                orientation="vertical",
            )
            return slider0, slider1

    def plot_connecting_points_update(self, val: tuple):
        """Event handler for the joint connection points sliders"""
        num_of_factors = self.plotted['mechanism'].factorizations[0].number_of_factors

        for i in range(num_of_factors):
            self.plotted['mechanism'].factorizations[0].linkage[i].set_point_by_param(0, self.joint_sliders[2 * i].val)
            self.plotted['mechanism'].factorizations[0].linkage[i].set_point_by_param(1, self.joint_sliders[1 + 2 * i].val)

        for i in range(num_of_factors):
            self.plotted['mechanism'].factorizations[1].linkage[i].set_point_by_param(0, self.joint_sliders[2 * num_of_factors + 2 * i].val)
            self.plotted['mechanism'].factorizations[1].linkage[i].set_point_by_param(1, self.joint_sliders[2 * num_of_factors + 1 + 2 * i].val)

        # update the plot
        self.plot_slider_update(self.move_slider.val)

    def plot_slider_update(self, val: float, t_param: float = None):
        """Event handler for the joint angle slider"""
        if t_param is not None:
            t = t_param
        else:
            # t parametrization for the driving joint
            t = self.plotted['mechanism'].factorizations[0].joint_angle_to_t_param(val)

        # plot links
        links = (self.plotted['mechanism'].factorizations[0].direct_kinematics(t)
                 + self.plotted['mechanism'].factorizations[1].direct_kinematics(t)[::-1])
        links.insert(0, links[-1])

        x, y, z = zip(*[links[j] for j in range(len(links))])

        for i, line in enumerate(self.lines):
            line.set_data_3d([x[i], x[i+1]], [y[i], y[i+1]], [z[i], z[i+1]])

        """
        xyz_coordinates = [(100 * xi, 100 * yi, 100 * zi) for xi, yi, zi in zip(x, y, z)]

        # Save XYZ coordinates to a CSV file
        pts_file_path = "xyz.pts"

        # Write the XYZ coordinates to the .pts file
        with open(pts_file_path, 'w') as pts_file:
            for point in xyz_coordinates:
                x, y, z = point
                pts_file.write(f"{x} {y} {z}\n")
        """

        if self.show_tool:
            # plot tool
            # use last point of each factorization
            tool_triangle = ([self.plotted['mechanism'].factorizations[0].direct_kinematics(t)[-1]]
                             + [self.plotted['mechanism'].factorizations[1].direct_kinematics(t)[-1]])
            # get tool point
            tool = self.plotted['mechanism'].factorizations[0].direct_kinematics_of_tool(
                t, self.plotted['mechanism'].tool_frame.dq2point_via_matrix())
            # add tool point to tool triangle
            tool_triangle.insert(1, tool)

            x, y, z = zip(*[tool_triangle[j] for j in range(len(tool_triangle))])
            self.tool_plot.set_data_3d(x, y, z)

            # plot tool frame
            pose_dq = DualQuaternion(self.plotted['mechanism'].evaluate(t))
            pose_matrix = TransfMatrix(pose_dq.dq2matrix()) * TransfMatrix(self.plotted['mechanism'].tool_frame.dq2matrix())

            x_vec, y_vec, z_vec = pose_matrix.get_plot_data()

            # remove old frame (quiver has no update method)
            for pose_arrow in self.pose_frame:
                pose_arrow.remove()
            # plot new frame
            self.pose_frame = [self.ax.quiver(*vec, color=color,
                                              length=self.arrows_length)
                               for vec, color in zip([x_vec, y_vec, z_vec],
                                                     ["red", "green", "blue"])]

        self.update_limits(self.ax)

        # update the plot
        if not self.jupyter_notebook:
            self.fig.canvas.draw_idle()
            self.fig.canvas.update()
            self.fig.canvas.flush_events()

    def show(self):
        """
        Show the plot
        """
        self.update_limits(self.ax)
        plt.show()

    def update_limits(self, ax):
        """
        Update the limits of the plot

        :param ax: matplotlib axes
        """
        # Inner function to update the minimum and maximum values
        def update_min_max(data):
            # Update min and max for x and y axes
            self.min_x, self.max_x = (min(self.min_x, np.min(data[:, 0])),
                                      max(self.max_x, np.max(data[:, 0])))
            self.min_y, self.max_y = (min(self.min_y, np.min(data[:, 1])),
                                      max(self.max_y, np.max(data[:, 1])))
            # Update min and max for z-axis if present
            if data.shape[1] > 2:
                self.min_z, self.max_z = (min(self.min_z, np.min(data[:, 2])),
                                          max(self.max_z, np.max(data[:, 2])))

        # Iterate over all artists in the Axes3D object
        for artist in ax.get_children():
            # Handle 3D scatter plots
            if isinstance(artist, matplotlib.collections.PathCollection):
                update_min_max(np.array(artist._offsets3d).T)
            # Handle 3D line plots
            elif hasattr(artist, '_verts3d'):
                update_min_max(np.array(artist._verts3d).T)
            # Handle 3D quiver plots
            elif hasattr(artist, '_segments3d'):
                for segment in artist._segments3d:
                    update_min_max(np.array(segment))
            # Handle other collection types (like PolyCollection for polygons)
            elif isinstance(artist, matplotlib.collections.Collection):
                for path in artist.get_paths():
                    for polygon in path.to_polygons():
                        update_min_max(np.array(polygon))

        # Set the updated limits to the axes
        ax.set_xlim3d(float(self.min_x), float(self.max_x))
        ax.set_ylim3d(float(self.min_y), float(self.max_y))
        ax.set_zlim3d(float(self.min_z), float(self.max_z))
        ax.set_aspect("equal")




from copy import deepcopy

import sympy as sp

from .MiniBall import MiniBall
from .PointHomogeneous import PointHomogeneous
from .RationalCurve import RationalCurve


class RationalBezier(RationalCurve):
    """
    Class representing rational Bezier curves in n-dimensional space.

    :examples:

    .. testcode::

        # Create a rational Bezier curve from control points

        # part of Limancon of Pascal

        from rational_linkages import RationalBezier, PointHomogeneous
        import numpy as np


        control_points = [PointHomogeneous(np.array([4.,  0., -2.,  4.])),
                          PointHomogeneous(np.array([0.,  1., -2.,  0.])),
                          PointHomogeneous(np.array([1.33333333, 2.66666667, 0., 1.33333333])),
                          PointHomogeneous(np.array([0., 1., 2., 0.])),
                          PointHomogeneous(np.array([4., 0., 2., 4.]))]
        bezier_curve = RationalBezier(control_points)
    """

    def __init__(self, control_points: list[PointHomogeneous], reparametrization:
    bool = False):
        """
        Initializes a RationalBezier object with the provided control points.

        :param list[PointHomogeneous] control_points: control points of the curve
        """
        super().__init__(
            self.get_coeffs_from_control_points(control_points, reparametrization=reparametrization)
        )

        self.control_points = control_points

        # Calculate the bounding ball of the control points
        self.ball = MiniBall(self.control_points)

    def get_coeffs_from_control_points(
        self, control_points: list[PointHomogeneous], reparametrization) -> (
            list[sp.Poly]):
        """
        Calculate the coefficients of the parametric equations of the curve from
        the control points.

        :param control_points:
        :param mapped: bool - True if the curve is mapped to the [-1,1] interval,
            False keeps [0,1] interval
        :return: np.array - coefficients of the parametric equations of the curve
        """
        t = sp.Symbol("t")
        degree = len(control_points) - 1
        dimension = control_points[0].coordinates.size - 1

        # Calculate the Bernstein basis polynomials and construct the Bezier curve
        bernstein_basis = self.get_bernstein_polynomial_equations(
            t, reparametrization=reparametrization, degree=degree
        )
        bezier_curve = [0] * (dimension + 1)
        for i in range(degree + 1):
            bezier_curve += bernstein_basis[i] * control_points[i].array()

        # Convert the Bezier curve to a set of polynomials
        bezier_polynomials = [
            sp.Poly(bezier_curve[i], t) for i in range(dimension + 1)]
        return bezier_polynomials

    def get_plot_data(self, interval: tuple = (0, 1), steps: int = 50) -> tuple:
        """
        Get the data to plot the curve in 2D or 3D, based on the dimension of the curve

        :param interval: tuple - interval of the parameter t
        :param steps: int - number of discrete steps in the interval to plot the curve

        :return: tuple - (x, y, z) coordinates of the curve
        """
        # perform superclass coordinates acquisition (only the curve)
        x, y, z = super().get_plot_data(interval, steps)

        # plot the control points
        x_cp, y_cp, z_cp = zip(
            *[self.control_points[i].normalized_in_3d() for i in range(self.degree + 1)]
        )
        return x, y, z, x_cp, y_cp, z_cp

    def check_for_control_points_at_infinity(self):
        """
        Check if there is a control point at infinity

        :return: bool - True if there is a control point at infinity, False otherwise
        """
        return any(
            self.control_points[i].is_at_infinity for i in range(self.degree + 1)
        )

    def split_de_casteljau(self, t=0.5):
        """
        Split the curve at the given parameter value t

        :param t: float - parameter value to split the curve at
        :return: tuple - two new Bezier curves
        """
        control_points = deepcopy(self.control_points)

        left_curve = [control_points[0]]
        right_curve = [control_points[-1]]

        # Perform De Casteljau subdivision until only two points remain
        while len(control_points) > 1:
            new_points = []

            # Compute linear interpolations between adjacent control points
            for i in range(len(control_points) - 1):
                new_points.append(
                    control_points[i].linear_interpolation(control_points[i + 1], t)
                )

            # Append the first point of the new segment to the left curve
            left_curve.append(new_points[0])
            # Insert the last point of a new segment at the beginning of the right curve
            right_curve.insert(0, new_points[-1])

            # Update control points for the next iteration
            control_points = new_points

        return RationalBezier(left_curve), RationalBezier(right_curve)



"""
Classes in the Module:
    - Linkage: Represents the connection points on a joint.
    - PointsConnection: Operates the connection points for a given joint.
    - LineSegment: Represents the physical realization of a linkage.
"""
from typing import Union

import numpy as np

from .DualQuaternion import DualQuaternion
from .NormalizedLine import NormalizedLine
from .PointHomogeneous import PointHomogeneous


class Linkage:
    """
    Class for storing the connection points on a joint.

    The connection points are stored as a list of PointHomogeneous objects. The
    default parameter of the connection points is 0 (nearest points on the axes to the
    origin).

    :ivar NormalizedLine normalized_axis: The axis of the joint
    :ivar PointsConnection points: The connection points
    :ivar list default_connection_point: The default connection point
    """
    def __init__(self, axis: DualQuaternion, connection_points: list[PointHomogeneous]):
        """
        :param DualQuaternion axis: The axis of the joint
        :param PointHomogeneous connection_points: The default connection point (
            common perpendicular)
        """
        self.normalized_axis = NormalizedLine(axis.dq2screw())

        if len(connection_points) == 1:
            self.default_connection_point = [connection_points[0], connection_points[0]]
        elif len(connection_points) == 2:
            self.default_connection_point = connection_points
        else:
            raise ValueError("Connection points must be a list of 1 or 2 points")

        self.points = PointsConnection(self.default_connection_point)

        # The parameters of the connection points are 0 by default (nearest point on
        # the axis to the origin), if not set differently
        self._params = [self._get_point_param_on_line(self.default_connection_point[0]),
                        self._get_point_param_on_line(self.default_connection_point[1])]
        self.set_point_by_param(0, self._get_point_param_on_line(self.default_connection_point[0]))
        self.set_point_by_param(1, self._get_point_param_on_line(self.default_connection_point[1]))

    @property
    def points_params(self) -> list[float, float]:
        """
        Returns the parameter of the connection points.

        :return: The parameter of the connection points
        :rtype: list[float, float]
        """
        return self._params

    @points_params.setter
    def points_params(self, value: list[float, float]):
        """
        Sets the parameter of the connection points.

        :param list[float, float] value: The parameter of the connection points
        """
        self._params = value

        if not self._check_equal_points():
            self.points[0] = self._get_point_using_param(value[0])
            self.points[1] = self._get_point_using_param(value[1])
        else:
            self.points[0] = self._get_point_using_param(value[0])
            self.points[1] = self._get_point_using_param(value[1] + 0.0001)

    def __repr__(self):
        return f"{self.points}"

    def _get_point_param_on_line(self, point: PointHomogeneous) -> np.ndarray:
        """
        Gets the parameter of the connection point at the given index.
        """
        if self.normalized_axis.contains_point(point.normalized_in_3d()):
            return self.normalized_axis.get_point_param(point.normalized_in_3d())
        else:
            print("Axis: {}".format(self.normalized_axis))
            print("Point: {}".format(point))
            raise ValueError("Point is not on the axis")

    def _get_point_using_param(self, param: float) -> PointHomogeneous:
        """
        Sets the connection point at the given parameter.

        :param float param: The parameter

        :return: The connection point
        :rtype: PointHomogeneous
        """
        return PointHomogeneous.from_3d_point(self.normalized_axis.point_on_line(param))

    def set_point_by_param(self, idx: int, param: Union[float, np.ndarray]):
        """
        Sets the connection point at the given parameter.

        :param int idx: Index of the connection parameter on the joint, 0 or 1.
        :param Union[float, np.ndarray] param: line-parameter defining the point on the
            line (joint axis)
        """
        if idx == 0:
            if param == self.points_params[1]:
                self.points_params = [param, self.points_params[1] + 0.0001]
            else:
                self.points_params = [param, self.points_params[1]]
        elif idx == 1:
            if param == self.points_params[0]:
                self.points_params = [self.points_params[0], param + 0.0001]
            else:
                self.points_params = [self.points_params[0], param]
        else:
            raise IndexError("Index out of range")

    def _check_equal_points(self) -> bool:
        """
        Checks if the connection points are equal.

        :return: True if the connection points are equal, False otherwise
        :rtype: bool
        """
        return np.allclose(self.points[0].normalized_in_3d(),
                           self.points[1].normalized_in_3d())


class PointsConnection:
    """
    Class for storing the connection points on a joint.

    :ivar PointHomogeneous _connection_point0: The first connection point
    :ivar PointHomogeneous _connection_point1: The second connection point
    """
    def __init__(self, connection_point: list[PointHomogeneous]):
        """
        :param PointHomogeneous connection_point: The default connection point (common
            perpendicular)
        """
        self._connection_point0 = connection_point[0]
        self._connection_point1 = connection_point[1]

    def __repr__(self):
        return f"{[self._connection_point0, self._connection_point1]}"

    def __getitem__(self, idx: int) -> PointHomogeneous:
        if idx == 0:
            return self._connection_point0
        elif idx == 1:
            return self._connection_point1
        else:
            raise IndexError("Index out of range")

    def __setitem__(self, idx: int, value):
        if idx == 0:
            self._connection_point0 = value
        elif idx == 1:
            self._connection_point1 = value
        else:
            raise IndexError("Index out of range")

    def __iter__(self):
        return iter([self._connection_point0, self._connection_point1])

    def __len__(self):
        return 2


class LineSegment:
    """
    Class for storing the physical realization of a linkage as their motion equations.

    :ivar NormalizedLine equation: The equation of the line segment under the motion
    :ivar PointHomogeneous point0: The equation of the first point of the line segment
    :ivar PointHomogeneous point1: The equation of the second point of the line segment
    :ivar str type: The type of the line segment (b - base, j - joint, l - link,
        t - tool)
    :ivar int factorization_idx: The index of the factorization the line segment
        belongs to
    :ivar int idx: The index of the line segment in the factorization
    """
    def __init__(self, eqation, point0, point1, linkage_type, f_idx, idx):
        self.equation = eqation
        self.point0 = point0
        self.point1 = point1
        self.type = linkage_type
        self.factorization_idx = f_idx
        self.idx = idx

    def __repr__(self):
        return f"{self.type}_{self.factorization_idx}{self.idx}"

    def is_point_in_segment(self, point: PointHomogeneous, t_val: float) -> bool:
        """
        Checks if the colliding point is in the line segment.

        :param PointHomogeneous point: The point
        :param float t_val: The parameter of collision point

        :return: True if the point is in the line segment, False otherwise
        :rtype: bool
        """
        # evaluate the connections points at the parameter t
        p0 = self.point0.evaluate(t_val)
        p1 = self.point1.evaluate(t_val)

        # segment length
        l = np.linalg.norm(p0.normalized_in_3d() - p1.normalized_in_3d())

        # distance between the point0 and the collision point
        d0 = np.linalg.norm(p0.normalized_in_3d() - point.normalized_in_3d())

        # distance between the point1 and the collision point
        d1 = np.linalg.norm(p1.normalized_in_3d() - point.normalized_in_3d())

        if np.allclose(l, d0 + d1):
            return True
        else:
            return False





import importlib.resources
import pickle

RationalMechanism = 'RationalMechanism'


def bennett_ark24() -> RationalMechanism:
    """
    Returns a RationalMechanism object of the Bennett linkage.

    This is a 4R linkage with 3 joints and 1 end effector and collision-free
    realization, that was introduced in the ARK 2024 conference paper: Rational
    Linkages: from Poses to 3D-printed Prototypes.

    :return: RationalMechanism object for the Bennett linkage.
    :rtype: RationalMechanism
    """
    resource_package = "rational_linkages.data"
    resource_path = 'bennett_ark24.pkl'
    with importlib.resources.path(resource_package, resource_path) as file_path:
        with open(file_path, 'rb') as f:
            return pickle.load(f)


def collisions_free_6r() -> RationalMechanism:
    """
    Returns a RationalMechanism object of a 6R collision-free realization.

    :return: RationalMechanism object for the 6R linkage.
    :rtype: RationalMechanism
    """
    resource_package = "rational_linkages.data"
    resource_path = 'collisions_free_6r.pkl'
    with importlib.resources.path(resource_package, resource_path) as file_path:
        with open(file_path, 'rb') as f:
            return pickle.load(f)


def plane_fold_6r() -> RationalMechanism:
    """
    Returns a RationalMechanism object of a 6R mechanism that folds in plane.

    Original model:
    h1 = DualQuaternion.as_rational([0, 1, 0, 0, 0, 0, 0, 0])
    h2 = DualQuaternion.as_rational([0, 0, 3, 0, 0, 0, 0, 1])
    h3 = DualQuaternion.as_rational([0, 1, 1, 0, 0, 0, 0, -2])

    :return: RationalMechanism object for the 6R linkage.
    :rtype: RationalMechanism
    """
    resource_package = "rational_linkages.data"
    resource_path = 'plane_fold_6r.pkl'
    with importlib.resources.path(resource_package, resource_path) as file_path:
        with open(file_path, 'rb') as f:
            return pickle.load(f)



import numpy as np


class AffineMatrix:
    """
    Class of affine matrices
    """
    def __init__(self, matrix: np.ndarray = None):
        """
        Construct the 4x4 affine matrix from the given matrix array

        :param matrix: np.ndarray - matrix
        """
        if matrix is None:
            matrix = np.eye(4)

        self.matrix = matrix

        self.vec_t = matrix[1:4, 0]  # translation vector
        self.vec_x = matrix[1:4, 1]
        self.vec_y = matrix[1:4, 2]
        self.vec_z = matrix[1:4, 3]

    @property
    def vector(self):
        """
        If the object is called, return 12-dimensional vector

        :return: np.ndarray - 12-dimensional vector of the matrix
        """
        return self.as_vector()

    def __repr__(self):
        return f"{self.vector}"

    def as_vector(self) -> np.ndarray:
        """
        Return the matrix as a 12D vector

        The vector is constructed by stacking the matrix in the following order:
        first column with translational vector, second column with x-axis vector,
        third column with y-axis vector, fourth column with z-axis vector.

        :return: np.ndarray - 12-dimensional vector of the matrix
        """
        return np.hstack((self.vec_t, self.vec_x, self.vec_y, self.vec_z))







import pickle
from copy import deepcopy
from time import time
from typing import Union

import numpy as np
import sympy as sp

from .DualQuaternion import DualQuaternion
from .MotionFactorization import MotionFactorization
from .NormalizedLine import NormalizedLine
from .RationalCurve import RationalCurve
from .TransfMatrix import TransfMatrix

PointHomogeneous = 'PointHomogeneous'


class RationalMechanism(RationalCurve):
    """
    Class representing rational mechanisms in dual quaternion space.

    :ivar factorizations: list of MotionFactorization objects
    :ivar num_joints: number of joints in the mechanism
    :ivar is_linkage: True if the mechanism is a linkage, False if it is 1 branch of a
        linkage
    :ivar tool_frame: end effector of the mechanism
    :ivar segments: list of LineSegment objects representing the physical realization of
        the linkage

    :examples:

    .. testcode::

        # Create a rational mechanism from given example

        from rational_linkages import RationalMechanism, Plotter
        from rational_linkages.models import bennett_ark24


        if __name__ == "__main__":
            # load the model of the Bennett's linkage
            m = bennett_ark24()

            # create an interactive plotter object, with 500 descrete steps
            # for the input rational curves, and arrows scaled to 0.05 length
            myplt = Plotter(interactive=True, steps=500, arrows_length=0.05)

            # plot the model with tool frame
            myplt.plot(m, show_tool=True)

            ##### additional plotting options #####
            # create a pose of the identity
            base = TransfMatrix()
            myplt.plot(base)

            # create another pose
            p0 = TransfMatrix.from_rpy_xyz([-90, 0, 0], [0.15, 0, 0], units='deg')
            myplt.plot(p0)
            ######################################

            # show the plot
            myplt.show()
    """

    def __init__(self, factorizations: list[MotionFactorization],
                 tool: Union[DualQuaternion, str] = None):
        """
        Initializes a RationalMechanism object
        """
        super().__init__(factorizations[0].set_of_polynomials)
        self.factorizations = factorizations
        self.num_joints = sum([f.number_of_factors for f in factorizations])

        self.is_linkage = True if len(self.factorizations) == 2 else False

        self.tool_frame = self._determine_tool(tool)

        if self.is_linkage:
            self.segments = self._get_line_segments_of_linkage()

    @classmethod
    def from_saved_file(cls, filename: str):
        """
        Load a linkage object from a file.

        :param str filename: name of the file to load the linkage object from

        :return: linkage object
        :rtype: RationalMechanism

        :raises FileNotFoundError: if the file was not possible to load
        """
        # check if the filename has the .pkl extension
        if filename[-4:] != '.pkl':
            filename = filename + '.pkl'

        try:
            with open(filename, 'rb') as file:
                mechanism = pickle.load(file)
        except FileNotFoundError:
            raise FileNotFoundError(f"File {filename} was not found or possible "
                                    f"to load.")

        return mechanism

    def save(self, filename: str = None):
        """
        Save the linkage object to a file.

        :param str filename: name of the file to save the linkage object to
        """
        if filename is None:
            filename = 'saved_mechanism.pkl'
        elif filename[-4:] == '.pkl':
            pass
        else:
            filename = filename + '.pkl'

        with open(filename, 'wb') as file:
            pickle.dump(self, file)

    def _determine_tool(self, tool: Union[DualQuaternion, None, str]) -> DualQuaternion:
        """
        Determine the tool frame of the mechanism.

        :return: tool frame of the mechanism
        :rtype: DualQuaternion
        """
        if tool is None:
            return DualQuaternion(self.evaluate(0, inverted_part=True))
        elif isinstance(tool, DualQuaternion):
            return tool
        elif tool == 'mid_of_last_link':
            # calculate the midpoint of the last link
            nearly_zero = np.finfo(float).eps
            p0 = self.factorizations[0].direct_kinematics(nearly_zero, inverted_part=True)[-1]
            p1 = self.factorizations[1].direct_kinematics(nearly_zero, inverted_part=True)[-1]

            # define the x axis vector - along the last link
            vec_x = (p1 - p0) / np.linalg.norm(p1 - p0)

            # get some random vector from the last joint points
            vec_y_pts = self.factorizations[0].direct_kinematics(nearly_zero, inverted_part=True)[-2:]
            vec_y = vec_y_pts[1] - vec_y_pts[0]

            # define the z axis vector - perpendicular to the x and y vectors
            vec_z = np.cross(vec_x, vec_y)
            vec_z = vec_z / np.linalg.norm(vec_z)

            mid = (p1 + p0) / 2

            t = TransfMatrix.from_vectors(normal_x=vec_x, approach_z=vec_z, origin=mid)
            return DualQuaternion(t.matrix2dq())
        else:
            raise ValueError("tool must be either DualQuaternion, "
                             "None default motion zero configuration, "
                             "or 'mid_of_last_link'")

    def get_design(self,
                   unit: str = 'rad',
                   scale: float = 1.0,
                   joint_length: float = 20.0,
                   washer_length: float = 1.0,
                   pretty_print: bool = True):
        """
        Get the design parameters of the linkage for the CAD model.

        The parameters are in the order: d, a, alpha, connection0,
        connection1, for every link.

        :param str unit: desired unit of the angle parameters, can be 'deg' or 'rad'
        :param float scale: scale of the length parameters of the linkage
        :param float joint_length: length of the joint segment in mm; default is 20 mm
            which corresponds to the CAD model that connects two 20 mm joint parts and
            has 1 mm thick washer between. Total length of the joint is 41 mm. It is
            used to calculate a midpoint distance between the two links that connect.
        :param float washer_length: length of the washer in mm; default is 1 mm
        :param bool pretty_print: if True, print the parameters in a readable form,
            otherwise return a numpy array

        :return: design parameters of the linkage
        :rtype: np.ndarray
        """
        screws = self.get_screw_axes()
        screws.append(screws[0])
        frames = self.get_frames()[1:]

        connection_params = self.get_segment_connections()
        midpts_dist = (joint_length + washer_length) / scale
        connection_params = self.map_connection_params(connection_params, midpts_dist)
        connection_params = np.vstack((connection_params, connection_params[0, :]))

        design_params = np.zeros((self.num_joints, 2))

        for i in range(self.num_joints):
            design_params[i, 0] = (connection_params[i, 1]
                                   - screws[i].get_point_param(frames[i].t))
            design_params[i, 1] = (connection_params[i+1, 0]
                                   - screws[i+1].get_point_param(frames[i+1].t))

        design_params = design_params * scale
        dh = self.get_dh_params(unit=unit, scale=scale)[1:]

        if pretty_print:
            for i in range(self.num_joints):
                print("---")
                print(f"Link {i}: d = {dh[i, 1]:.6f}, "
                      f"a = {dh[i, 2]:.6f}, "
                      f"alpha = {dh[i, 3]:.6f}")
                print(f"cp_0 = {design_params[i, 0]:.6f}, "
                      f"cp_1 = {design_params[i, 1]:.6f}")

        return dh, design_params

    def get_segment_connections(self) -> np.ndarray:
        """
        Get the connection parameters of the linkage.

        :return: connection points of the linkage
        :rtype: np.ndarray
        """
        connection_params = np.zeros((self.num_joints, 2))
        for i in range(len(self.factorizations[0].linkage)):
            connection_params[i, :] = self.factorizations[0].linkage[i].points_params

        for i in range(len(self.factorizations[1].linkage)):
            # iterate from back to front
            connection_params[-1-i, :] = self.factorizations[1].linkage[i].points_params[::-1]

        return connection_params

    def get_dh_params(self, unit: str = 'rad', scale: float = 1.0) -> np.ndarray:
        """
        Get the standard Denavit-Hartenberg parameters of the linkage.

        The parameters are in the order: theta, d, a, alpha. It follows the standard
        convention. The first row is are the parameters of the base frame.

        See more in the paper by :footcite:t:`Huczala2022iccma`.

        :param str unit: desired unit of the angle parameters, can be 'deg' or 'rad'
        :param float scale: scale of the length parameters of the linkage

        :return: theta, d, a, alpha array of Denavit-Hartenberg parameters
        :rtype: np.ndarray

        .. footbibliography::

        """
        frames = self.get_frames()

        # closed-loop mechanism - add 1st joint at the end of the list
        frames.append(frames[1])

        dh = np.zeros((self.num_joints + 1, 4))
        for i in range(self.num_joints + 1):
            th, d, a, al = frames[i].dh_to_other_frame(frames[i+1])

            if unit == 'deg':
                th = np.rad2deg(th)
                al = np.rad2deg(al)
            elif unit != 'rad':
                raise ValueError("unit must be deg or rad")

            dh[i, :] = [th, scale * d, scale * a, al]
        return dh

    def get_frames(self) -> list[TransfMatrix]:
        """
        Get the frames of a linkage that follow standard Denaivt-Hartenberg convention.

        It renurns n+2 frames, where n is the number of joints. The first frame is the
        base frame, and the last frame is an updated frame of the first joint that
        follows the DH convention in respect to the last joint's frame.

        :return: list of TransfMatrix objects
        :rtype: list[TransfMatrix]
        """
        from .TransfMatrix import TransfMatrix

        frames = [TransfMatrix()] * (self.num_joints + 2)

        screws = self.get_screw_axes()

        # add the first screw to the end of the list
        screws.append(screws[0])

        # insert origin as the base line
        screws.insert(0, NormalizedLine())

        for i, line in enumerate(screws[1:]):
            # obtain the connection points and the distance to the previous line
            pts, dist, cos_angle = line.common_perpendicular_to_other_line(screws[i])
            vec = pts[0] - pts[1]

            if not np.isclose(dist, 0.0):  # if the lines are skew or parallel
                # normalize vec - future X axis
                vec_x = vec / np.linalg.norm(vec)

                # from line.dir (future Z axis) and x create an SE3 object
                frames[i+1] = TransfMatrix.from_vectors(vec_x,
                                                        line.direction,
                                                        origin=pts[0])

            else:  # Z axes are intersecting or coincident
                if np.isclose(np.dot(frames[i].a, line.direction), 1):
                    # Z axes are coincident, therefore the frames are the same
                    frames[i+1] = deepcopy(frames[i])

                elif np.isclose(np.dot(frames[i].a, line.direction), -1):
                    # Z axes are coincident, but differ in orientation
                    rot_x_pi = TransfMatrix.from_rpy([np.pi, 0, 0])
                    frames[i + 1] = TransfMatrix(frames[i].matrix @ rot_x_pi.matrix)

                else:  # Z axis intersect with an angle
                    # future X axis as cross product of previous Z axis and new Z axis
                    vec_x = np.cross(frames[i].a, line.direction)

                    frames[i + 1] = TransfMatrix.from_vectors(vec_x,
                                                              line.direction,
                                                              origin=pts[0])

        return frames

    def get_screw_axes(self) -> list[NormalizedLine]:
        """
        Get the normalized lines (screw axes, Plucker coordinates) of the linkage.

        The lines are in home configuration. They consist of two factorizations, and
        the second factorization axes must be reversed.

        :return: list of NormalizedLine objects
        :rtype: list[NormalizedLine]
        """
        screws = []
        for axis in self.factorizations[0].dq_axes:
            screws.append(NormalizedLine(axis.dq2screw()))

        branch2 = []
        for axis in self.factorizations[1].dq_axes:
            branch2.append(NormalizedLine(axis.dq2screw()))

        return screws + branch2[::-1]

    def map_connection_params(self,
                              connection_params: np.ndarray,
                              midpoints_distance: float) -> np.ndarray:
        """
        Map the connection parameters to the given joint length.

        :param np.ndarray connection_params: list of connection points parameters of the
            linkage
        :param float midpoints_distance: distance between the midpoints of the two links
            that connect at a joint

        :return: mapped connection points parameters
        :rtype: np.ndarray
        """
        c_params = np.asarray(connection_params)

        for i in range(len(c_params)):
            c_params_len = np.linalg.norm(c_params[i, 0] - c_params[i, 1])

            if not np.allclose(c_params_len, midpoints_distance):
                c_params[i, :] = self._map_joint_segment(c_params[i, :],
                                                         midpoints_distance)

        return c_params

    @staticmethod
    def _map_joint_segment(points_params: np.ndarray, midpoints_distance: float):
        """
        Map the joint segment to the scale of the linkage.

        :param np.ndarray points_params: list of connection points parameters of the
            joint segment

        :param float midpoints_distance: distance between the midpoints of the two links
            that connect at a joint

        :return: mapped joint segment
        :rtype: np.ndarray
        """

        def map_interval(input_interval, max_length):
            mid_point = (input_interval[0] + input_interval[1]) / 2

            if input_interval[0] < input_interval[1]:
                mapped_interval = [mid_point - max_length/2, mid_point + max_length/2]
            else:
                mapped_interval = [mid_point + max_length/2, mid_point - max_length/2]
            return mapped_interval

        points_params = np.asarray(points_params)
        new_params = map_interval(points_params, midpoints_distance)

        # Calculate midpoints
        #midpoint1 = new_params[0] + (new_params[1] - new_params[0]) / 4
        #midpoint2 = new_params[0] + 3 * (new_params[1] - new_params[0]) / 4

        return new_params
    
    def collision_check(self,
                        parallel: bool = False,
                        pretty_print: bool = True,
                        only_links: bool = False,
                        terminate_on_first: bool = False):
        """
        Perform full-cycle collision check on the line-model linkage.

        By default, the collision check is performed in non-parallel mode. This is
        faster for 4-bar linkages and 6-bar lingakes with a "simpler" motion curve,
        but slower for 6-bar linkages with "complex" motions.

        :param bool parallel: if True, perform collision check in parallel using
            multiprocessing
        :param bool pretty_print: if True, print the results in a readable form
        :param bool only_links: if True, only link-link collisions are checked,
            expecting that distances between joint connection points are minimal
        :param bool terminate_on_first: if True, terminate the collision check when
            the first collision is found

        :return: list of collision check colliding parameter values
        :rtype: list[float]
        """
        start_time = time()
        print("Collision check started...")

        # update the line segments (physical realization of the linkage)
        self.segments = self._get_line_segments_of_linkage()

        iters = []
        # iterate over all line segments
        for ii in range(len(self.segments)):
            # for each line segment, iterate over all other line segments that are not
            # its immediate neighbors
            for jj in range(ii + 2, len(self.segments)):
                # in only links should be checked (joint segments have minimal length)
                if only_links:
                    if (self.segments[ii].type == 'j'
                            or self.segments[jj].type == 'j'
                            or jj - ii == 2):  # skip neighbouring links
                        pass
                    else:
                        iters.append((ii, jj))
                else:
                    iters.append((ii, jj))

        # remove the last neighbouring pair of links
        if only_links:
            # find the pair with the highest difference
            max_pair = max(iters, key=lambda x: x[1] - x[0])
            # remove the pair from the list
            iters.remove(max_pair)

        print(f"--- number of tasks to solve: {len(iters)} ---")

        if parallel:
            collision_results = self._collision_check_parallel(iters)
        else:
            collision_results = self._collision_check_nonparallel(iters,
                                                                  terminate_on_first)

        results = [r for r in collision_results if r is not None]
        flattened_results = [item for sublist in results for item in sublist]
        if len(results) == 0:
            flattened_results = None

        end_time = time()
        print(f"--- collision check finished in {end_time - start_time} seconds.")

        if pretty_print:
            if flattened_results is None:
                print("No collisions found.")
            else:
                print(f"The linkage is colliding {len(flattened_results)} times at the "
                      f"following parameter values:")
                for res in flattened_results:
                    print(res)

        return flattened_results

    def _collision_check_parallel(self, iters: list[tuple[int, int]]):
        """
        Perform collision check in parallel using multiprocessing.

        Slower for 4-bar linkages and 6-bar lingakes with a "simpler" motion curve,
        faster for 6-bar linkages with "complex" motions.

        :param list iters: list of tuples of indices of the line segments to be checked

        :return: list of collision check results
        :rtype: list[str]
        """
        print("--- running in parallel ---")
        import concurrent.futures

        with concurrent.futures.ProcessPoolExecutor() as executor:
            results = executor.map(self._check_given_pair, iters)

        return list(results)

    def _collision_check_nonparallel(self, iters: list[tuple[int, int]],
                                     terminate_on_first: bool = False):
        """
        Perform collision check in non-parallel mode.

        Default option. Faster for 4-bar linkages and 6-bar lingakes with a "simpler"
        motion curve, slower for 6-bar linkages with "complex" motions.

        :param list iters: list of tuples of indices of the line segments to be checked
        :param bool terminate_on_first: if True, terminate the collision check when
            the first collision is found

        :return: list of collision check results
        :rtype: list[str]
        """
        results = []
        for val in iters:
            collsion = self._check_given_pair(val)
            results.append(collsion)
            if terminate_on_first and collsion is not None:
                break
        return results

    def _check_given_pair(self, iters: tuple[int, int]) -> list[float]:
        """
        Perform collision check for a given pair of line segments and evaluate it.

        :param tuple iters: tuple of indices of the line segments to be checked

        :return: collision check result
        :rtype: list[float]
        """
        i = iters[0]
        j = iters[1]
        # check if two lines are colliding
        collisions, coll_pts = self.colliding_lines(self.segments[i].equation,
                                                    self.segments[j].equation)

        if collisions is not None:
            # check if the collision is between the physical line segments
            physical_collision = [
                self.segments[i].is_point_in_segment(coll_pts[k], t_val) and
                self.segments[j].is_point_in_segment(coll_pts[k], t_val)
                for k, t_val in enumerate(collisions)
            ]
        else:
            physical_collision = [False]

        result = []
        if True in physical_collision:
            for i, res in enumerate(physical_collision):
                if res:
                    result.append(collisions[i])
            #result = f"{physical_collision} at parameters: {collisions} for linkage pair: {self.segments[i].type}_{self.segments[i].factorization_idx}{self.segments[i].idx} X {self.segments[j].type}_{self.segments[j].factorization_idx}{self.segments[j].idx}"
        else:
            #result = f"no collision between {self.segments[i].type}_{self.segments[i].factorization_idx}{self.segments[i].idx} X {self.segments[j].type}_{self.segments[j].factorization_idx}{self.segments[j].idx}"
            result = None

        return result

    def colliding_lines(self, l0: NormalizedLine, l1: NormalizedLine
                        ) -> tuple[list[float], list[PointHomogeneous]]:
        """
        Return the lines that are colliding in the linkage.

        :param NormalizedLine l0: equation of the first line
        :param NormalizedLine l1: equation of the second line

        :return: tuple (list of t values, list of intersection points)
        :rtype: tuple[list[float], list[PointHomogeneous]]
        """
        t = sp.Symbol("t")

        # lines are colliding if expr == 0
        expr = sp.simplify(np.dot(l0.direction, l1.moment) + np.dot(l0.moment, l1.direction))

        # neighbouring lines are colliding all the time (expr == 0)
        if expr == 0:
            return None, None

        e = sp.Expr(expr).subs(t, (t + 1) / 2)

        expr_poly = sp.Poly(e.args[0], t)

        expr_coeffs = expr_poly.all_coeffs()

        # convert to numpy polynomial
        expr_n = np.array(expr_coeffs, dtype="float64")
        np_poly = np.polynomial.polynomial.Polynomial(expr_n[::-1])
        np_poly_inversed = np.polynomial.polynomial.Polynomial(expr_n)

        # solve for t
        colliding_lines_sol = np_poly.roots()
        colliding_lines_sol_inversed = np_poly_inversed.roots()
        # extract real solutions
        sol_real = colliding_lines_sol.real[np.isclose(colliding_lines_sol.imag,
                                                       0, atol=1e-5)]
        sol_real_inversed = colliding_lines_sol_inversed.real[
            np.isclose(colliding_lines_sol_inversed.imag, 0, atol=1e-5)]

        sol_real = [((sol + 1)/2) for sol in sol_real]
        sol_real_inversed = [((sol + 1)/2) for sol in sol_real_inversed]

        solutions = deepcopy(sol_real)

        for sol in sol_real_inversed:
            if np.isclose(sol, 0):
                # eps is very small number (avoid division by zero)
                sol = 1 / np.finfo(float).eps
            else:
                sol = 1 / sol

            solutions = np.append(solutions, sol)

        intersection_points = self.get_intersection_points(l0, l1, solutions)

        return solutions, intersection_points

    def get_intersection_points(self, l0: NormalizedLine, l1: NormalizedLine,
                                t_params: list[float]):
        """
        Return the intersection points of two lines.

        :param NormalizedLine l0: equation of the first line
        :param NormalizedLine l1: equation of the second line
        :param list[float] t_params: list of parameter values - points of intersection

        :return: list of intersection points
        :rtype: list[PointHomogeneous]
        """
        from .PointHomogeneous import PointHomogeneous
        intersection_points = [PointHomogeneous()] * len(t_params)

        for i, t_val in enumerate(t_params):
            # evaluate the lines at the given parameter
            l0e = l0.evaluate(t_val)
            l1e = l1.evaluate(t_val)

            # common perpendicular to the two lines - there is none since they
            # intersect, therefore from the list of two points only 1 is needed
            inters_points, d, c = l0e.common_perpendicular_to_other_line(l1e)
            intersection_points[i] = PointHomogeneous.from_3d_point(inters_points[0])

        return intersection_points

    def _get_line_segments_of_linkage(self) -> list:
        """
        Return the line segments of the linkage.

        Line segments are the physical realization of the linkage. This method obtains
        their motion equations using default connection points of the factorizations
        (default meaning the static points in the home configuration).

        :return: list of LineSegment objects
        :rtype: list[LineSegment]
        """
        from .Linkage import LineSegment

        t = sp.Symbol("t")

        segments = [[], []]

        # base (static) link has index 0 in the list of the 1st factorization
        eq, p0, p1 = self.factorizations[0].base_link(self.factorizations[1].linkage[0].points[0])
        segments[0].append(LineSegment(eq, p0, p1, linkage_type="b", f_idx=0, idx=0))

        # static joints
        segments[0].append(LineSegment(*self.factorizations[0].joint(0),
                                       linkage_type="j", f_idx=0, idx=0))
        segments[1].append(LineSegment(*self.factorizations[1].joint(0),
                                       linkage_type="j", f_idx=1, idx=0))

        # moving links and joints
        for i in range(2):
            for j in range(1, self.factorizations[i].number_of_factors):
                link, p0, p1 = self.factorizations[i].link(j)
                link = self.factorizations[i].act(link, end_idx=j-1, param=t)
                p0 = self.factorizations[i].act(p0, end_idx=j-1, param=t)
                p1 = self.factorizations[i].act(p1, end_idx=j-1, param=t)
                segments[i].append(LineSegment(link, p0, p1, linkage_type="l", f_idx=i, idx=j))

                joint, p0, p1 = self.factorizations[i].joint(j)
                joint = self.factorizations[i].act(joint, end_idx=j, param=t)
                p0 = self.factorizations[i].act(p0, end_idx=j, param=t)
                p1 = self.factorizations[i].act(p1, end_idx=j, param=t)
                segments[i].append(LineSegment(joint, p0, p1, linkage_type="j", f_idx=i, idx=j))

        # tool (moving - acted) link has index -1 in the list of the 2nd factorization
        tool_link, p0, p1 = self.factorizations[0].tool_link(self.factorizations[1].linkage[-1].points[1])
        tool_link = self.factorizations[0].act(tool_link, param=t)
        p0 = self.factorizations[0].act(p0, param=t)
        p1 = self.factorizations[1].act(p1, param=t)
        tool_idx = self.factorizations[1].number_of_factors
        segments[1].append(LineSegment(tool_link, p0, p1, linkage_type="t", f_idx=1, idx=tool_idx))

        return segments[0] + segments[1][::-1]

    def get_motion_curve(self):
        """
        Return the rational motion curve of the linkage as RationalCurve object.

        :return: motion curve of the linkage
        :rtype: RationalCurve
        """
        return self.curve()

    def singularity_check(self):
        """
        Perform singularity check of the mechanism.
        """
        from .SingularityAnalysis import SingularityAnalysis

        sa = SingularityAnalysis()
        return sa.check_singularity(self)

    def smallest_polyline(self, update_design: bool = False) -> tuple[list, list, float]:
        """
        Obtain the smallest polyline links for the mechanism.

        :param bool update_design: if True, update also the design of the mechanism

        :return: list of points of the smallest polyline, list of points parameters,
            result of the optimization
        :rtype: list, list, float
        """
        from .CollisionFreeOptimization import CollisionFreeOptimization

        # get smallest polyline
        pts, points_params, res = CollisionFreeOptimization(self).smallest_polyline()

        # update the design of the mechanism
        if update_design:
            self.factorizations[0].set_joint_connection_points_by_parameters(
                points_params[:len(self.factorizations[0].dq_axes)])
            self.factorizations[1].set_joint_connection_points_by_parameters(
                points_params[len(self.factorizations[1].dq_axes):][::-1])

        return pts, points_params, res

    def collision_free_optimization(self,
                                    method: str = None,
                                    step_length=25,
                                    min_joint_segment_length: float = 0.001,
                                    max_iters: int = 10,
                                    **kwargs):
        """
        Perform collision-free optimization of the mechanism.

        :param str method: method of optimization, can be 'combinatorial_search' by
            :footcite:t:`Li2020`
        :param float step_length: length of the step, i.e. the shift distance value, see
            :ref:`combinatorial_search` for more detail
        :param float min_joint_segment_length: minimum length of the joint segment
        :param int max_iters: maximum number of iterations
        :param kwargs: additional keyword arguments

        :return: list of collision-free points parameters
        :rtype: list
        """
        from .CollisionFreeOptimization import CollisionFreeOptimization
        optimizer = CollisionFreeOptimization(self)

        if method is None:
            method = 'combinatorial_search'

        match method:
            case 'combinatorial_search':
                results = optimizer.optimize(method=method,
                                             step_length=step_length,
                                             min_joint_segment_length=min_joint_segment_length,
                                             max_iters=max_iters,
                                             **kwargs)
            case _:
                raise ValueError("Invalid method.")

        return results



import numpy as np
import sympy as sp

from .DualQuaternion import DualQuaternion

# Forward declarations for class names
NormalizedLine = "NormalizedLine"
PointHomogeneous = "PointHomogeneous"


class RationalDualQuaternion(DualQuaternion):
    """
    RationalDualQuaternion class representing a 8-dimensional dual quaternion.
    """
    def __init__(self, study_parameters: list[sp.Rational]):
        """
        RationalDualQuaternion class

        :param study_parameters: 8-vector list of dual quaternion parameters
        """
        self.rational_numbers = study_parameters

        floating_point_numbers = np.asarray(study_parameters)
        super().__init__(floating_point_numbers)

        self.is_rational = True

    def __repr__(self):
        """
        Printing method override

        :return: Rational Dual Quaterion in readable form
        :rtype: str
        """
        return f"{self.rational_numbers}"

    def __getitem__(self, idx) -> sp.Rational:
        """
        Get an element of DualQuaternion

        :param int idx: index of the Quaternion element to call 0..7

        :return: float number of the element
        :rtype: np.ndarray
        """
        return self.rational_numbers[idx]

    def array(self):
        """
        Get the array of the rational numbers

        :return: Rational numbers
        :rtype: sp.Matrix
        """
        return np.array(self.rational_numbers)
